from pathlib import Path

from trilogy.core.models.environment import Environment, LazyEnvironment


def test_multi_environment():
    basic = Environment()

    basic.parse(
        """
const pi <- 3.14;
                     

""",
        namespace="math",
    )

    basic.parse(
        """
                
            select math.pi;
                """
    )

    assert basic.concepts["math.pi"].name == "pi"


def test_test_alias_free_import():
    basic = Environment(working_path=Path(__file__).parent)

    basic.parse(
        """
import test_env;

key id2 int;


""",
    )

    assert basic.concepts["id"].name == "id"
    assert basic.concepts["id2"].name == "id2"
    assert basic.concepts["id"].namespace == basic.concepts["id2"].namespace


def test_import_concept_resolution():
    basic = LazyEnvironment(
        load_path=Path(__file__).parent / "test_lazy_env.preql",
        working_path=Path(__file__).parent,
        setup_queries=[],
    )
    materialized = basic.materialize_for_select()
    assert "one.two.import_key" in materialized.materialized_concepts
    assert "two.two.import_key" in materialized.materialized_concepts


def test_import_basics():
    basic = Environment(working_path=Path(__file__).parent)

    basic.parse(
        """
import test_env;

key id2 int;


""",
    )

    assert len(basic.imports["local"]) == 1, basic.imports
    importz = basic.imports["local"][0]
    assert importz.path == Path("test_env")
    expected = Path(__file__).parent / "test_env.preql"
    assert importz.input_path == expected



from trilogy import Environment
from trilogy.ai.conversation import Conversation
from trilogy.ai.enums import Provider
from trilogy.ai.providers.base import LLMProvider


def text_to_query(
    environment: Environment,
    user_input: str,
    provider: Provider,
    model: str,
    secret: str | None = None,
) -> str:
    llm_provider: LLMProvider

    if provider == Provider.OPENAI:
        from trilogy.ai.providers.openai import OpenAIProvider

        llm_provider = OpenAIProvider(
            name="openai",
            api_key=secret,
            model=model,
        )
    elif provider == Provider.ANTHROPIC:
        from trilogy.ai.providers.anthropic import AnthropicProvider

        llm_provider = AnthropicProvider(
            name="anthropic",
            api_key=secret,
            model=model,
        )
    elif provider == Provider.GOOGLE:
        from trilogy.ai.providers.google import GoogleProvider

        llm_provider = GoogleProvider(
            name="google",
            api_key=secret,
            model=model,
        )
    else:
        raise ValueError(f"Unsupported provider: {provider}")
    conversation = Conversation.create(
        provider=llm_provider,
    )

    response = conversation.generate_query(
        user_input=user_input, environment=environment
    )

    return response



from typing import List

from trilogy.constants import logger
from trilogy.core.models.build import (
    BuildConcept,
    BuildFilterItem,
    BuildWhereClause,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.processing.node_generators.common import (
    resolve_filter_parent_concepts,
)
from trilogy.core.processing.nodes import (
    FilterNode,
    History,
    MergeNode,
    SelectNode,
    StrategyNode,
)
from trilogy.core.processing.utility import is_scalar_condition, padding, unique

LOGGER_PREFIX = "[GEN_FILTER_NODE]"

FILTER_TYPES = (BuildFilterItem,)


def pushdown_filter_to_parent(
    local_optional: List[BuildConcept],
    conditions: BuildWhereClause | None,
    filter_where: BuildWhereClause,
    same_filter_optional: list[BuildConcept],
    depth: int,
) -> bool:
    optimized_pushdown = False
    if not is_scalar_condition(filter_where.conditional):
        optimized_pushdown = False
    elif not local_optional:
        optimized_pushdown = True
    elif conditions and conditions == filter_where:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} query conditions are the same as filter conditions, can optimize across all concepts"
        )
        optimized_pushdown = True
    elif same_filter_optional == local_optional:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} all optional concepts are included in the filter, can optimize across all concepts"
        )
        optimized_pushdown = True

    return optimized_pushdown


def build_parent_concepts(
    concept: BuildConcept,
    environment: BuildEnvironment,
    local_optional: List[BuildConcept],
    conditions: BuildWhereClause | None = None,
    depth: int = 0,
) -> tuple[
    list[BuildConcept],
    list[tuple[BuildConcept, ...]],
    list[BuildConcept],
    bool,
    bool,
]:
    parent_row_concepts, parent_existence_concepts = resolve_filter_parent_concepts(
        concept, environment
    )
    if not isinstance(concept.lineage, FILTER_TYPES):
        raise SyntaxError('Filter node must have a filter type lineage"')
    filter_where = concept.lineage.where

    same_filter_optional: list[BuildConcept] = []
    # mypy struggled here? we shouldn't need explicit bools
    global_filter_is_local_filter: bool = (
        True if (conditions and conditions == filter_where) else False
    )

    exact_partial_matches = True
    for x in local_optional:
        if isinstance(x.lineage, FILTER_TYPES):
            if set([x.address for x in x.lineage.where.concept_arguments]) == set(
                [x.address for x in filter_where.concept_arguments]
            ):
                exact_partial_matches = (
                    exact_partial_matches and x.lineage.where == filter_where
                )
                logger.info(
                    f"{padding(depth)}{LOGGER_PREFIX} fetching parents for peer {x.address} (of {concept.address})"
                )

                for arg in x.lineage.content_concept_arguments:
                    if arg.address not in parent_row_concepts:
                        parent_row_concepts.append(arg)
                same_filter_optional.append(x)
                continue
        elif global_filter_is_local_filter:
            same_filter_optional.append(x)
            # also append it to the parent row concepts
            parent_row_concepts.append(x)

    # sometimes, it's okay to include other local optional above the filter
    # in case it is, prep our list
    extra_row_level_optional: list[BuildConcept] = []

    for x in local_optional:
        if x.address in same_filter_optional:
            continue
        extra_row_level_optional.append(x)
    is_optimized_pushdown = exact_partial_matches and pushdown_filter_to_parent(
        local_optional, conditions, filter_where, same_filter_optional, depth
    )
    if not is_optimized_pushdown:
        parent_row_concepts += extra_row_level_optional
    return (
        parent_row_concepts,
        parent_existence_concepts,
        same_filter_optional,
        is_optimized_pushdown,
        global_filter_is_local_filter,
    )


def add_existence_sources(
    core_parent_nodes: list[StrategyNode],
    parent_existence_concepts: list[tuple[BuildConcept, ...]],
    source_concepts,
    environment,
    g,
    depth,
    history,
):
    for existence_tuple in parent_existence_concepts:
        if not existence_tuple:
            continue
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} fetching filter node existence parents {[x.address for x in existence_tuple]}"
        )
        parent_existence = source_concepts(
            mandatory_list=list(existence_tuple),
            environment=environment,
            g=g,
            depth=depth + 1,
            history=history,
        )
        if not parent_existence:
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} filter existence node parents could not be found"
            )
            return None
        core_parent_nodes.append(parent_existence)


def gen_filter_node(
    concept: BuildConcept,
    local_optional: List[BuildConcept],
    environment: BuildEnvironment,
    g,
    depth: int,
    source_concepts,
    history: History | None = None,
    conditions: BuildWhereClause | None = None,
) -> StrategyNode | None:
    if not isinstance(concept.lineage, FILTER_TYPES):
        raise SyntaxError('Filter node must have a filter type lineage"')
    where = concept.lineage.where

    (
        parent_row_concepts,
        parent_existence_concepts,
        same_filter_optional,
        optimized_pushdown,
        global_filter_is_local_filter,
    ) = build_parent_concepts(
        concept,
        environment=environment,
        local_optional=local_optional,
        conditions=conditions,
        depth=depth,
    )

    row_parent: StrategyNode = source_concepts(
        mandatory_list=parent_row_concepts,
        environment=environment,
        g=g,
        depth=depth + 1,
        history=history,
        conditions=conditions,
    )

    core_parent_nodes: list[StrategyNode] = []
    flattened_existence = [x for y in parent_existence_concepts for x in y]
    if parent_existence_concepts:
        add_existence_sources(
            core_parent_nodes,
            parent_existence_concepts,
            source_concepts,
            environment,
            g,
            depth,
            history,
        )

    if not row_parent:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} filter node row parents {[x.address for x in parent_row_concepts]} could not be found"
        )
        return None
    else:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} filter node has row parents {[x.address for x in parent_row_concepts]} from node with output [{[x.address for x in row_parent.output_concepts]}] partial {row_parent.partial_concepts}"
        )
    if global_filter_is_local_filter:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} filter node conditions match global conditions adding row parent {row_parent.output_concepts} with condition {where.conditional}"
        )
        row_parent.add_parents(core_parent_nodes)
        # all local optional will be in the parent already, so we can set outputs
        row_parent.set_output_concepts([concept] + local_optional)
        return row_parent
    if optimized_pushdown:
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} returning optimized filter node with pushdown to parent with condition {where.conditional} across {[concept] + same_filter_optional + row_parent.output_concepts} "
        )
        if isinstance(row_parent, SelectNode):
            logger.info(
                f"{padding(depth)}{LOGGER_PREFIX} nesting select node in strategy node"
            )
            parent = StrategyNode(
                input_concepts=row_parent.output_concepts,
                output_concepts=[concept]
                + same_filter_optional
                + row_parent.output_concepts,
                environment=row_parent.environment,
                parents=[row_parent],
                depth=row_parent.depth,
                partial_concepts=row_parent.partial_concepts,
                force_group=False,
            )
        else:
            parent = row_parent
            parent.add_output_concepts([concept] + same_filter_optional)
        parent.add_parents(core_parent_nodes)
        if not parent.preexisting_conditions == where.conditional:
            parent.add_condition(where.conditional)
        parent.add_existence_concepts(flattened_existence, False)
        # parent.grain = BuildGrain.from_concepts(
        #     parent.output_concepts,
        #     environment=environment,
        # )
        parent.rebuild_cache()
        filter_node = parent
    else:
        core_parent_nodes.append(row_parent)
        filter_node = FilterNode(
            input_concepts=unique(
                parent_row_concepts + flattened_existence,
                "address",
            ),
            output_concepts=[concept] + same_filter_optional + parent_row_concepts,
            environment=environment,
            parents=core_parent_nodes,
            preexisting_conditions=conditions.conditional if conditions else None,
        )

    if not local_optional or all(
        [x.address in filter_node.output_concepts for x in local_optional]
    ):
        optional_outputs = [
            x for x in filter_node.output_concepts if x.address in local_optional
        ]
        logger.info(
            f"{padding(depth)}{LOGGER_PREFIX} no extra enrichment needed for filter node, has all of {[x.address for x in local_optional]}"
        )
        filter_node.set_output_concepts(
            [
                concept,
            ]
            + optional_outputs
        )
        return filter_node
    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} need to enrich filter node with additional concepts {[x.address for x in local_optional if x.address not in filter_node.output_concepts]}"
    )
    enrich_node: StrategyNode = source_concepts(  # this fetches the parent + join keys
        # to then connect to the rest of the query
        mandatory_list=parent_row_concepts
        + [x for x in local_optional if x.address not in filter_node.output_concepts],
        environment=environment,
        g=g,
        depth=depth + 1,
        history=history,
        conditions=conditions,
    )
    if not enrich_node:
        logger.error(
            f"{padding(depth)}{LOGGER_PREFIX} filter node enrichment node could not be found"
        )
        return filter_node
    logger.info(
        f"{padding(depth)}{LOGGER_PREFIX} returning filter node and enrich node with {enrich_node.output_concepts} and {enrich_node.input_concepts}"
    )
    return MergeNode(
        input_concepts=filter_node.output_concepts + enrich_node.output_concepts,
        output_concepts=[
            concept,
        ]
        + local_optional,
        environment=environment,
        parents=[
            filter_node,
            enrich_node,
        ],
        preexisting_conditions=conditions.conditional if conditions else None,
    )



from enum import Enum
from logging import DEBUG, StreamHandler
from typing import Union
from uuid import uuid4

from trilogy.constants import logger
from trilogy.core.models.build import BuildDatasource
from trilogy.core.models.execute import (
    CTE,
    QueryDatasource,
    UnionCTE,
)
from trilogy.core.processing.nodes import StrategyNode
from trilogy.core.statements.author import SelectStatement
from trilogy.hooks.base_hook import BaseHook


class PrintMode(Enum):
    OFF = False
    BASIC = True
    FULL = 3


class DebuggingHook(BaseHook):
    def __init__(
        self,
        level=DEBUG,
        max_depth: int | None = None,
        process_ctes: PrintMode | bool = True,
        process_nodes: PrintMode | bool = True,
        process_datasources: PrintMode | bool = True,
        process_other: bool = True,
    ):
        if not any([isinstance(x, StreamHandler) for x in logger.handlers]):
            logger.addHandler(StreamHandler())
        logger.setLevel(level)

        self.max_depth = max_depth
        self.process_ctes = PrintMode(process_ctes)
        self.process_nodes = PrintMode(process_nodes)
        self.process_datasources = PrintMode(process_datasources)
        self.process_other = PrintMode(process_other)
        self.messages: list[str] = []
        self.uuid = uuid4()
        from trilogy.dialect.bigquery import BigqueryDialect

        self.renderer = BigqueryDialect()

    def print(self, *args):
        merged = " ".join([str(x) for x in args])
        self.messages.append(merged)

    def write(self):
        with open(f"debug_{self.uuid}.log", "w") as f:
            f.write("\n".join(self.messages))

    def process_select_info(self, select: SelectStatement):
        if self.process_datasources != PrintMode.OFF:
            self.print(f"grain: {str(select.grain)}")

    def process_root_datasource(self, datasource: QueryDatasource):
        if self.process_datasources != PrintMode.OFF:
            printed = self.print_recursive_resolved(
                datasource, self.process_datasources
            )
            for row in printed:
                self.print("".join([str(v) for v in row]))

    def process_root_cte(self, cte: CTE | UnionCTE):
        if self.process_ctes != PrintMode.OFF:
            self.print_recursive_ctes(cte, max_depth=self.max_depth)

    def process_root_strategy_node(self, node: StrategyNode):
        if self.process_nodes != PrintMode.OFF:
            printed = self.print_recursive_nodes(node, mode=self.process_nodes)
            for row in printed:
                # logger.info("".join([str(v) for v in row]))
                self.print("".join([str(v) for v in row]))

    def print_recursive_resolved(
        self,
        input: Union[QueryDatasource, BuildDatasource],
        mode: PrintMode,
        depth: int = 0,
    ):
        extra = []
        if isinstance(input, QueryDatasource):
            if input.joins:
                extra.append("join")
            if input.condition:
                extra.append("filter")
        if input.group_required:
            extra.append("group")
        output = [c.address for c in input.output_concepts[:3]]
        if len(input.output_concepts) > 3:
            output.append("...")
        display = [
            (
                "  " * depth,
                input.__class__.__name__,
                "<",
                ",".join(extra),
                ">",
                # [c.address for c in input.input_concepts],
                "->",
                output,
            )
        ]
        if isinstance(input, QueryDatasource):
            for child in input.datasources:
                display += self.print_recursive_resolved(
                    child, mode=mode, depth=depth + 1
                )
        return display

    def print_recursive_ctes(
        self, input: CTE | UnionCTE, depth: int = 0, max_depth: int | None = None
    ):
        if max_depth and depth > max_depth:
            return
        select_statement = [c.address for c in input.output_columns]
        self.print(
            "  " * depth, input.name, "->", input.group_to_grain, "->", select_statement
        )
        sql = self.renderer.render_cte(input).statement
        for line in sql.split("\n"):
            logger.debug("  " * (depth) + line)
        if isinstance(input, CTE):
            for child in input.parent_ctes:
                self.print_recursive_ctes(child, depth + 1)
        elif isinstance(input, UnionCTE):
            for child in input.parent_ctes:
                for parent in child.parent_ctes:
                    self.print_recursive_ctes(parent, depth + 1)

    def print_recursive_nodes(
        self, input: StrategyNode, mode: PrintMode = PrintMode.BASIC, depth: int = 0
    ):
        resolved = input.resolve()
        if mode == PrintMode.FULL:
            display = [
                [
                    "  " * depth,
                    input,
                    "->",
                    resolved.grain,
                    "->",
                    [c.address for c in resolved.output_concepts],
                ]
            ]
        elif mode == PrintMode.BASIC:
            display = [
                [
                    "  " * depth,
                    input,
                    "->",
                    resolved.grain,
                ]
            ]
        for child in input.parents:
            display += self.print_recursive_nodes(
                child,
                mode=mode,
                depth=depth + 1,
            )
        return display



from trilogy.constants import CONFIG, logger
from trilogy.core.enums import BooleanOperator, Derivation
from trilogy.core.models.build import (
    BuildConditional,
)
from trilogy.core.models.execute import CTE, RecursiveCTE, UnionCTE
from trilogy.core.optimizations import (
    HideUnusedConcepts,
    InlineDatasource,
    OptimizationRule,
    PredicatePushdown,
    PredicatePushdownRemove,
)
from trilogy.core.processing.utility import sort_select_output
from trilogy.core.statements.author import MultiSelectStatement, SelectStatement

MAX_OPTIMIZATION_LOOPS = 100


# other optimizations may make a CTE a pure passthrough
# remove those
# def is_locally_irrelevant(cte: CTE) -> CTE | bool:
#     if not len(cte.parent_ctes) == 1:
#         return False
#     parent = cte.parent_ctes[0]
#     if not parent.output_columns == cte.output_columns:
#         return False
#     if cte.condition is not None:
#         return False
#     if cte.group_to_grain:
#         return False
#     if len(cte.joins)>1:
#         return False
#     return parent


def reorder_ctes(
    input: list[CTE],
):
    import networkx as nx

    # Create a directed graph
    G = nx.DiGraph()
    mapping: dict[str, CTE] = {}
    for cte in input:
        mapping[cte.name] = cte
        for parent in cte.parent_ctes:
            G.add_edge(parent.name, cte.name)
    # Perform topological sort (only works for DAGs)
    try:
        topological_order = list(nx.topological_sort(G))
        if not topological_order:
            return input
        return [mapping[x] for x in topological_order]
    except nx.NetworkXUnfeasible as e:
        logger.error(
            "The graph is not a DAG (contains cycles) and cannot be topologically sorted."
        )
        raise e


def filter_irrelevant_ctes(
    input: list[CTE | UnionCTE],
    root_cte: CTE | UnionCTE,
):
    relevant_ctes = set()

    def recurse(cte: CTE | UnionCTE, inverse_map: dict[str, list[CTE | UnionCTE]]):
        # TODO: revisit this
        # if parent := is_locally_irrelevant(cte):
        #     logger.info(
        #         f"[Optimization][Irrelevent CTE filtering] Removing redundant CTE {cte.name} and replacing with {parent.name}"
        #     )
        #     for child in inverse_map.get(cte.name, []):
        #         child.parent_ctes = [
        #             x for x in child.parent_ctes if x.name != cte.name
        #         ] + [parent]
        #         for x in child.source_map:
        #             if cte.name in child.source_map[x]:
        #                 child.source_map[x].remove(cte.name)
        #                 child.source_map[x].append(parent.name)
        #         for x2 in child.existence_source_map:
        #             if cte.name in child.existence_source_map[x2]:
        #                 child.existence_source_map[x2].remove(cte.name)
        #                 child.existence_source_map[x2].append(parent.name)
        # else:
        relevant_ctes.add(cte.name)

        for parent in cte.parent_ctes:
            if parent.name in relevant_ctes:
                logger.info(
                    f"[Optimization][Irrelevent CTE filtering] Already visited {parent.name} when visting {cte.name}, potential recursive dag"
                )
                continue

            recurse(parent, inverse_map)
        if isinstance(cte, UnionCTE):
            for internal in cte.internal_ctes:
                recurse(internal, inverse_map)

    inverse_map = gen_inverse_map(input)
    recurse(root_cte, inverse_map)
    final = [cte for cte in input if cte.name in relevant_ctes]
    filtered = [cte for cte in input if cte.name not in relevant_ctes]
    if filtered:
        logger.info(
            f"[Optimization][Irrelevent CTE filtering] Removing redundant CTEs {[x.name for x in filtered]}"
        )
    if len(final) == len(input):
        return input
    return filter_irrelevant_ctes(final, root_cte)


def gen_inverse_map(input: list[CTE | UnionCTE]) -> dict[str, list[CTE | UnionCTE]]:
    inverse_map: dict[str, list[CTE | UnionCTE]] = {}
    for cte in input:
        if isinstance(cte, UnionCTE):
            for internal in cte.internal_ctes:
                if internal.name not in inverse_map:
                    inverse_map[internal.name] = []
                inverse_map[internal.name].append(cte)
        else:
            for parent in cte.parent_ctes:
                if parent.name not in inverse_map:
                    inverse_map[parent.name] = []
                inverse_map[parent.name].append(cte)

    return inverse_map


SENSITIVE_DERIVATIONS = [
    Derivation.UNNEST,
    Derivation.WINDOW,
    Derivation.RECURSIVE,
]


def is_direct_return_eligible(cte: CTE | UnionCTE) -> CTE | UnionCTE | None:
    # if isinstance(select, (PersistStatement, MultiSelectStatement)):
    #     return False
    if len(cte.parent_ctes) != 1:
        return None
    direct_parent = cte.parent_ctes[0]
    if isinstance(direct_parent, (UnionCTE, RecursiveCTE)):
        return None

    output_addresses = set([x.address for x in cte.output_columns])
    parent_output_addresses = set([x.address for x in direct_parent.output_columns])
    if not output_addresses.issubset(parent_output_addresses):
        return None
    if not direct_parent.grain == cte.grain:
        logger.info("[Direct Return] grain mismatch, cannot early exit")
        return None

    assert isinstance(cte, CTE)
    derived_concepts = [
        c for c in cte.source.output_concepts if c not in cte.source.input_concepts
    ]

    parent_derived_concepts = [
        c
        for c in direct_parent.source.output_concepts
        if c not in direct_parent.source.input_concepts
    ]
    condition_arguments = cte.condition.row_arguments if cte.condition else []
    for x in derived_concepts:
        if x.derivation in SENSITIVE_DERIVATIONS:
            return None
    for x in parent_derived_concepts:
        if x.address not in condition_arguments:
            continue
        if x.derivation in SENSITIVE_DERIVATIONS:
            return None
    for x in condition_arguments:
        # if it's derived in the parent
        if x.address in parent_derived_concepts:
            if x.derivation in SENSITIVE_DERIVATIONS:
                return None
            # this maybe needs to be recursive if we flatten a ton of derivation
            # into one CTE
            if not x.lineage:
                continue
            for z in x.lineage.concept_arguments:
                # if it was preexisting in the parent, it's safe
                if z.address in direct_parent.source.input_concepts:
                    continue
                # otherwise if it's dangerous, play it safe.
                if z.derivation in SENSITIVE_DERIVATIONS:
                    return None
    logger.info(
        f"[Optimization][EarlyReturn] Removing redundant output CTE {cte.name} with derived_concepts {[x.address for x in derived_concepts]}"
    )
    return direct_parent


def optimize_ctes(
    input: list[CTE | UnionCTE],
    root_cte: CTE | UnionCTE,
    select: SelectStatement | MultiSelectStatement,
) -> list[CTE | UnionCTE]:
    direct_parent: CTE | UnionCTE | None = root_cte
    while CONFIG.optimizations.direct_return and (
        direct_parent := is_direct_return_eligible(root_cte)
    ):
        direct_parent.order_by = root_cte.order_by
        direct_parent.limit = root_cte.limit
        direct_parent.hidden_concepts = root_cte.hidden_concepts.union(
            direct_parent.hidden_concepts
        )
        if root_cte.condition:
            if direct_parent.condition:
                direct_parent.condition = BuildConditional(
                    left=direct_parent.condition,
                    operator=BooleanOperator.AND,
                    right=root_cte.condition,
                )
            else:
                direct_parent.condition = root_cte.condition
        root_cte = direct_parent

        sort_select_output(root_cte, select)

    REGISTERED_RULES: list["OptimizationRule"] = []

    if CONFIG.optimizations.datasource_inlining:
        REGISTERED_RULES.append(InlineDatasource())
    if CONFIG.optimizations.predicate_pushdown:
        REGISTERED_RULES.append(PredicatePushdown())
    if CONFIG.optimizations.predicate_pushdown:
        REGISTERED_RULES.append(PredicatePushdownRemove())
    if CONFIG.optimizations.hide_unused_concepts:
        REGISTERED_RULES.append(HideUnusedConcepts())
    for rule in REGISTERED_RULES:
        loops = 0
        complete = False
        while not complete and (loops <= MAX_OPTIMIZATION_LOOPS):
            actions_taken = False
            # assume we go through all CTEs once
            look_at = [root_cte, *reversed(input)]
            inverse_map = gen_inverse_map(look_at)
            for cte in look_at:
                opt = rule.optimize(cte, inverse_map)
                actions_taken = actions_taken or opt
            complete = not actions_taken
            loops += 1
        input = reorder_ctes(filter_irrelevant_ctes(input, root_cte))
        logger.info(
            f"[Optimization] Finished checking for {type(rule).__name__} after {loops} loop(s)"
        )

    return reorder_ctes(filter_irrelevant_ctes(input, root_cte))



import itertools
from collections import defaultdict
from typing import List

from trilogy.constants import logger
from trilogy.core.enums import Derivation
from trilogy.core.models.build import BuildConcept, BuildWhereClause
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.processing.nodes import History, StrategyNode
from trilogy.core.processing.utility import padding

LOGGER_PREFIX = "[GEN_SYNONYM_NODE]"


def gen_synonym_node(
    all_concepts: List[BuildConcept],
    environment: BuildEnvironment,
    g,
    depth: int,
    source_concepts,
    history: History | None = None,
    conditions: BuildWhereClause | None = None,
    accept_partial: bool = False,
) -> StrategyNode | None:
    local_prefix = f"{padding(depth)}[GEN_SYNONYM_NODE]"
    base_fingerprint = tuple(sorted([x.address for x in all_concepts]))
    synonyms = defaultdict(list)
    has_synonyms = False
    for x in all_concepts:
        synonyms[x.address] = [x]
        if x.address in environment.alias_origin_lookup:
            parent = environment.concepts[x.address]
            if parent.address != x.address:
                synonyms[x.address].append(parent)
                has_synonyms = True
        for y in x.pseudonyms:
            if y in environment.alias_origin_lookup:
                synonyms[x.address].append(environment.alias_origin_lookup[y])
                has_synonyms = True
            elif y in environment.concepts:
                synonyms[x.address].append(environment.concepts[y])
                has_synonyms = True
    for address in synonyms:
        synonyms[address].sort(key=lambda obj: obj.address)
    if not has_synonyms:
        return None

    logger.info(f"{local_prefix} Generating Synonym Node with {len(synonyms)} synonyms")
    sorted_keys = sorted(synonyms.keys())
    combinations_list: list[tuple[BuildConcept, ...]] = list(
        itertools.product(*(synonyms[obj] for obj in sorted_keys))
    )

    def similarity_sort_key(combo: tuple[BuildConcept, ...]):
        addresses = [x.address for x in combo]

        # Calculate similarity score - count how many pairs share prefixes
        similarity_score = 0
        roots = sum(
            [1 for x in combo if x.derivation in (Derivation.ROOT, Derivation.CONSTANT)]
        )
        for i in range(len(addresses)):
            for j in range(i + 1, len(addresses)):
                # Find common prefix length
                addr1_parts = addresses[i].split(".")
                addr2_parts = addresses[j].split(".")
                common_prefix_len = 0
                for k in range(min(len(addr1_parts), len(addr2_parts))):
                    if addr1_parts[k] == addr2_parts[k]:
                        common_prefix_len += 1
                    else:
                        break
                similarity_score += common_prefix_len

        # Sort by roots, similarity (descending), then by addresses (ascending) for ties
        return (-roots, -similarity_score, addresses)

    combinations_list.sort(key=similarity_sort_key)
    for combo in combinations_list:
        fingerprint = tuple(sorted([x.address for x in combo]))
        if fingerprint == base_fingerprint:
            continue
        logger.info(
            f"{local_prefix} checking combination {fingerprint} with {len(combo)} concepts"
        )
        attempt: StrategyNode | None = source_concepts(
            list(combo),
            history=history,
            environment=environment,
            depth=depth,
            conditions=conditions,
            g=g,
            accept_partial=accept_partial,
        )
        if attempt:
            logger.info(f"{local_prefix} found inputs with {combo}")
            return attempt
    return None



from typing import Any, Callable, Mapping

from jinja2 import Template

from trilogy.core.enums import FunctionType, UnnestMode, WindowType
from trilogy.dialect.base import BaseDialect

ENV_SNOWFLAKE_PW = "PREQL_SNOWFLAKE_PW"
ENV_SNOWFLAKE_USER = "PREQL_SNOWFLAKE_USER"
ENV_SNOWFLAKE_ACCOUNT = "PREQL_SNOWFLAKE_ACCOUNT"

WINDOW_FUNCTION_MAP: Mapping[WindowType, Callable[[Any, Any, Any], str]] = {}

FUNCTION_MAP = {
    FunctionType.COUNT: lambda x, types: f"count({x[0]})",
    FunctionType.SUM: lambda x, types: f"sum({x[0]})",
    FunctionType.LENGTH: lambda x, types: f"length({x[0]})",
    FunctionType.AVG: lambda x, types: f"avg({x[0]})",
    FunctionType.LIKE: lambda x, types: (
        f" CASE WHEN {x[0]} like {x[1]} THEN True ELSE False END"
    ),
    FunctionType.MINUTE: lambda x, types: f"EXTRACT(MINUTE from {x[0]})",
    FunctionType.SECOND: lambda x, types: f"EXTRACT(SECOND from {x[0]})",
    FunctionType.HOUR: lambda x, types: f"EXTRACT(HOUR from {x[0]})",
    FunctionType.DAY_OF_WEEK: lambda x, types: f"EXTRACT(DAYOFWEEK from {x[0]})",
    FunctionType.DAY: lambda x, types: f"EXTRACT(DAY from {x[0]})",
    FunctionType.YEAR: lambda x, types: f"EXTRACT(YEAR from {x[0]})",
    FunctionType.MONTH: lambda x, types: f"EXTRACT(MONTH from {x[0]})",
    FunctionType.WEEK: lambda x, types: f"EXTRACT(WEEK from {x[0]})",
    FunctionType.QUARTER: lambda x, types: f"EXTRACT(QUARTER from {x[0]})",
    # math
    FunctionType.DIVIDE: lambda x, types: f"DIV0({x[0]},{x[1]})",
    FunctionType.UNNEST: lambda x, types: f"table(flatten({x[0]}))",
    FunctionType.ARRAY: lambda x, types: f"ARRAY_CONSTRUCT({', '.join(x)})",
}

FUNCTION_GRAIN_MATCH_MAP = {
    **FUNCTION_MAP,
    FunctionType.COUNT_DISTINCT: lambda args, types: f"CASE WHEN{args[0]} IS NOT NULL THEN 1 ELSE 0 END",
    FunctionType.COUNT: lambda args, types: f"CASE WHEN {args[0]} IS NOT NULL THEN 1 ELSE 0 END",
    FunctionType.SUM: lambda args, types: f"{args[0]}",
    FunctionType.AVG: lambda args, types: f"{args[0]}",
}


SNOWFLAKE_SQL_TEMPLATE = Template(
    """{%- if output %}
CREATE OR REPLACE TABLE {{ output.address.location }} AS
{% endif %}{%- if ctes %}
WITH {% if recursive%}RECURSIVE{% endif %}{% for cte in ctes %}
"{{cte.name}}" as ({{cte.statement}}){% if not loop.last %},{% endif %}{% else %}
{% endfor %}{% endif %}
{%- if full_select -%}
{{full_select}}
{%- else -%}

SELECT
{%- for select in select_columns %}
    {{ select }}{% if not loop.last %},{% endif %}{% endfor %}
{% if base %}FROM
    {{ base }}{% endif %}{% if joins %}{% for join in joins %}
    {{ join }}{% endfor %}{% endif %}
{% if where %}WHERE
    {{ where }}
{% endif %}
{%- if group_by %}GROUP BY {% for group in group_by %}
    {{group}}{% if not loop.last %},{% endif %}{% endfor %}{% endif %}{% if having %}
HAVING
\t{{ having }}{% endif %}
{%- if order_by %}
ORDER BY {% for order in order_by %}
    {{ order }}{% if not loop.last %},{% endif %}{% endfor %}{% endif %}
{%- if limit is not none %}
LIMIT {{ limit }}{% endif %}{% endif %}
"""
)
MAX_IDENTIFIER_LENGTH = 50


class SnowflakeDialect(BaseDialect):
    WINDOW_FUNCTION_MAP = {**BaseDialect.WINDOW_FUNCTION_MAP, **WINDOW_FUNCTION_MAP}
    FUNCTION_MAP = {**BaseDialect.FUNCTION_MAP, **FUNCTION_MAP}
    FUNCTION_GRAIN_MATCH_MAP = {
        **BaseDialect.FUNCTION_GRAIN_MATCH_MAP,
        **FUNCTION_GRAIN_MATCH_MAP,
    }
    QUOTE_CHARACTER = '"'
    SQL_TEMPLATE = SNOWFLAKE_SQL_TEMPLATE
    UNNEST_MODE = UnnestMode.SNOWFLAKE



def safe_quote(string: str, quote_char: str):
    # split dotted identifiers
    # TODO: evaluate if we need smarter parsing for strings that could actually include .
    if string.startswith("https://"):
        # it's a url, no splitting
        return f"{quote_char}{string}{quote_char}"
    components = string.split(".")
    return ".".join([f"{quote_char}{string}{quote_char}" for string in components])



from os import environ
from typing import Any, Dict, List, Optional

from trilogy.ai.enums import Provider
from trilogy.ai.models import LLMMessage, LLMResponse, UsageDict

from .base import LLMProvider, LLMRequestOptions
from .utils import RetryOptions, fetch_with_retry


class GoogleProvider(LLMProvider):
    def __init__(
        self,
        name: str,
        model: str,
        api_key: str | None = None,
        retry_options: Optional[RetryOptions] = None,
    ):
        api_key = api_key or environ.get("GOOGLE_API_KEY")
        if not api_key:
            raise ValueError(
                "API key argument or environment variable GOOGLE_API_KEY is required"
            )
        super().__init__(name, api_key, model, Provider.GOOGLE)
        self.base_model_url = "https://generativelanguage.googleapis.com/v1/models"
        self.base_completion_url = "https://generativelanguage.googleapis.com/v1beta"
        self.models: List[str] = []
        self.type = Provider.GOOGLE
        self.retry_options = retry_options or RetryOptions(
            max_retries=3,
            initial_delay_ms=30000,  # 30s default for Google's 429 rate limits
            retry_status_codes=[429, 500, 502, 503, 504],
            on_retry=lambda attempt, delay_ms, error: print(
                f"Google API retry attempt {attempt} after {delay_ms}ms delay due to error: {str(error)}"
            ),
        )

    def _convert_to_gemini_history(
        self, messages: List[LLMMessage]
    ) -> List[Dict[str, Any]]:
        """Convert standard message format to Gemini format."""
        return [
            {
                "role": "model" if msg.role == "assistant" else "user",
                "parts": [{"text": msg.content}],
            }
            for msg in messages
        ]

    def generate_completion(
        self, options: LLMRequestOptions, history: List[LLMMessage]
    ) -> LLMResponse:
        try:
            import httpx
        except ImportError:
            raise ImportError(
                "Missing httpx. Install pytrilogy[ai] to use GoogleProvider."
            )

        # Convert messages to Gemini format
        gemini_history = self._convert_to_gemini_history(history)

        # Separate system message if present
        system_instruction = None
        contents = gemini_history

        # Check if first message is a system message
        if history and history[0].role == "system":
            system_instruction = {"parts": [{"text": history[0].content}]}
            contents = gemini_history[1:]  # Remove system message from history

        # Build the request URL
        url = f"{self.base_completion_url}/models/{self.model}:generateContent"

        # Build request body
        request_body: Dict[str, Any] = {"contents": contents, "generationConfig": {}}

        # Add system instruction if present
        if system_instruction:
            request_body["systemInstruction"] = system_instruction

        # Add generation config options
        if options.temperature is not None:
            request_body["generationConfig"]["temperature"] = options.temperature

        if options.max_tokens is not None:
            request_body["generationConfig"]["maxOutputTokens"] = options.max_tokens

        if options.top_p is not None:
            request_body["generationConfig"]["topP"] = options.top_p

        try:
            # Make the API request with retry logic using a lambda
            response = fetch_with_retry(
                fetch_fn=lambda: httpx.post(
                    url,
                    headers={
                        "Content-Type": "application/json",
                        "x-goog-api-key": self.api_key,
                    },
                    json=request_body,
                    timeout=60.0,
                ),
                options=self.retry_options,
            )

            response.raise_for_status()
            data = response.json()

            # Extract text from response
            candidates = data.get("candidates", [])
            if not candidates:
                raise Exception("No candidates returned from Google API")

            content = candidates[0].get("content", {})
            parts = content.get("parts", [])

            if not parts:
                raise Exception("No parts in response content")

            text = parts[0].get("text", "")

            # Extract usage metadata
            usage_metadata = data.get("usageMetadata", {})
            prompt_tokens = usage_metadata.get("promptTokenCount", 0)
            completion_tokens = usage_metadata.get("candidatesTokenCount", 0)

            return LLMResponse(
                text=text,
                usage=UsageDict(
                    prompt_tokens=prompt_tokens,
                    completion_tokens=completion_tokens,
                    total_tokens=prompt_tokens + completion_tokens,
                ),
            )
        except httpx.HTTPStatusError as error:
            error_detail = error.response.text
            raise Exception(
                f"Google API error ({error.response.status_code}): {error_detail}"
            )
        except Exception as error:
            raise Exception(f"Google API error: {str(error)}")



from trilogy import Environment, Executor
from trilogy.core.enums import Derivation, Modifier, Purpose
from trilogy.core.exceptions import (
    ConceptModelValidationError,
    DatasourceColumnBindingData,
    DatasourceColumnBindingError,
)
from trilogy.core.models.build import (
    BuildConcept,
)
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.validation.common import ExpectationType, ValidationTest, easy_query


def validate_property_concept(
    concept: BuildConcept, exec: Executor | None = None
) -> list[ValidationTest]:
    return []


def validate_key_concept(
    concept: BuildConcept,
    env: Environment,
    build_env: BuildEnvironment,
    exec: Executor | None = None,
):
    results: list[ValidationTest] = []
    seen: dict[str, int] = {}

    count = 0
    for datasource in build_env.datasources.values():
        if concept.address in [c.address for c in datasource.concepts]:
            count += 1
    # if it only has one source, it's a key
    if count <= 1:
        return results

    for datasource in build_env.datasources.values():
        if concept.address in [c.address for c in datasource.concepts]:
            assignment = [
                x for x in datasource.columns if x.concept.address == concept.address
            ][0]
            # if it's not a partial, skip it
            if not assignment.is_complete:
                continue
            type_query = easy_query(
                concepts=[
                    # build_env.concepts[concept.address],
                    build_env.concepts[f"grain_check_{concept.safe_address}"],
                ],
                datasource=datasource,
                env=env,
                limit=1,
            )
            if exec:
                type_sql = exec.generate_sql(type_query)[-1]

                rows = exec.execute_raw_sql(type_sql).fetchall()
                seen[datasource.name] = rows[0][0] if rows else 0
            else:
                results.append(
                    ValidationTest(
                        raw_query=type_query,
                        check_type=ExpectationType.ROWCOUNT,
                        expected=f"equal_max_{concept.safe_address}",
                        result=None,
                        ran=False,
                    )
                )

    if not exec:
        return results
    max_seen: int = max([v for v in seen.values() if v is not None], default=0)
    for datasource in build_env.datasources.values():
        if concept.address in [c.address for c in datasource.concepts]:
            assignment = [
                x for x in datasource.columns if x.concept.address == concept.address
            ][0]
            err = None
            datasource_count: int = seen.get(datasource.name, 0)
            if datasource_count < max_seen and assignment.is_complete:
                err = DatasourceColumnBindingError(
                    address=datasource.identifier,
                    errors=[
                        DatasourceColumnBindingData(
                            address=concept.address,
                            value=None,
                            value_type=concept.datatype,
                            value_modifiers=[Modifier.PARTIAL],
                            actual_type=concept.datatype,
                            actual_modifiers=concept.modifiers,
                        )
                    ],
                    message=f"Key concept {concept.address} is missing values in datasource {datasource.name} (max cardinality in data {max_seen}, datasource has {seen[datasource.name]} values) but is not marked as partial.",
                )
            results.append(
                ValidationTest(
                    check_type=ExpectationType.ROWCOUNT,
                    expected=str(max_seen),
                    result=err,
                    ran=True,
                )
            )

    return results


def validate_datasources(
    concept: BuildConcept, build_env: BuildEnvironment
) -> list[ValidationTest]:
    if concept.lineage:
        return []
    for datasource in build_env.datasources.values():
        if concept.address in [c.address for c in datasource.concepts]:
            return []
    if not concept.derivation == Derivation.ROOT:
        return []
    if concept.name.startswith("__") or (
        concept.namespace and concept.namespace.startswith("__")
    ):
        return []
    return [
        ValidationTest(
            check_type=ExpectationType.LOGICAL,
            expected=None,
            result=ConceptModelValidationError(
                f"Concept {concept.address} is a root concept but has no datasources bound"
            ),
            ran=True,
        )
    ]


def validate_concept(
    concept: BuildConcept,
    env: Environment,
    build_env: BuildEnvironment,
    exec: Executor | None = None,
) -> list[ValidationTest]:
    base: list[ValidationTest] = []
    base += validate_datasources(concept, build_env)
    if concept.purpose == Purpose.PROPERTY:
        base += validate_property_concept(concept)
    elif concept.purpose == Purpose.KEY:
        base += validate_key_concept(concept, env, build_env, exec)
    return base



from __future__ import annotations

import hashlib
from abc import ABC
from datetime import date, datetime
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    List,
    Optional,
    Self,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationInfo,
    computed_field,
    field_validator,
    model_validator,
)

from trilogy.constants import DEFAULT_NAMESPACE, MagicConstants
from trilogy.core.constants import ALL_ROWS_CONCEPT
from trilogy.core.enums import (
    BooleanOperator,
    ComparisonOperator,
    ConceptSource,
    DatePart,
    Derivation,
    FunctionClass,
    FunctionType,
    Granularity,
    InfiniteFunctionArgs,
    Modifier,
    Ordering,
    Purpose,
    WindowOrder,
    WindowType,
)
from trilogy.core.models.core import (
    Addressable,
    ArrayType,
    DataType,
    DataTyped,
    ListWrapper,
    MapType,
    MapWrapper,
    NumericType,
    StructType,
    TraitDataType,
    TupleWrapper,
    arg_to_datatype,
    is_compatible_datatype,
)
from trilogy.utility import unique

# TODO: refactor to avoid these
if TYPE_CHECKING:
    from trilogy.core.models.environment import Environment


class Namespaced(ABC):
    def with_namespace(self, namespace: str):
        raise NotImplementedError


class Mergeable(ABC):
    def with_merge(self, source: Concept, target: Concept, modifiers: List[Modifier]):
        raise NotImplementedError

    def with_reference_replacement(self, source: str, target: Expr | ArgBinding):
        raise NotImplementedError(type(self))


class ConceptArgs(ABC):
    @property
    def concept_arguments(self) -> Sequence["ConceptRef"]:
        raise NotImplementedError

    @property
    def existence_arguments(self) -> Sequence[tuple["ConceptRef", ...]]:
        return []

    @property
    def row_arguments(self) -> Sequence["ConceptRef"]:
        return self.concept_arguments


class HasUUID(ABC):
    @property
    def uuid(self) -> str:
        return hashlib.md5(str(self).encode()).hexdigest()


class ConceptRef(Addressable, Namespaced, DataTyped, Mergeable, BaseModel):
    address: str
    datatype: (
        DataType | TraitDataType | ArrayType | StructType | MapType | NumericType
    ) = DataType.UNKNOWN
    metadata: Optional["Metadata"] = None

    @property
    def reference(self):
        return self

    @property
    def line_no(self) -> int | None:
        if self.metadata:
            return self.metadata.line_number
        return None

    def __repr__(self):
        return f"ref:{self.address}"

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if isinstance(other, Concept):
            return self.address == other.address
        elif isinstance(other, str):
            return self.address == other
        elif isinstance(other, ConceptRef):
            return self.address == other.address
        return False

    @property
    def namespace(self):
        return self.address.rsplit(".", 1)[0]

    @property
    def name(self):
        return self.address.rsplit(".", 1)[1]

    @property
    def output_datatype(self):
        return self.datatype

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> ConceptRef:
        if self.address == source.address:
            return ConceptRef.model_construct(
                address=target.address, datatype=target.datatype, metadata=self.metadata
            )
        return self

    def with_namespace(self, namespace: str):
        return ConceptRef.model_construct(
            address=address_with_namespace(self.address, namespace),
            datatype=self.datatype,
            metadata=self.metadata,
        )

    def with_reference_replacement(self, source: str, target: Expr | ArgBinding):
        # a reference might be to an attribute of a struct that is bound late
        # if the replacement is a parent in the access path; replace reference
        # with an attribute access call
        candidates = [f"{DEFAULT_NAMESPACE}.{self.address}", self.address]
        for candidate in candidates:
            if candidate == source:
                return target
            if not candidate.startswith(f"{source}."):
                continue
            attribute = self.address.rsplit(".", 1)[1]
            dtype = arg_to_datatype(target)
            if not isinstance(dtype, StructType):
                continue
            output_type = dtype.field_types.get(attribute, DataType.UNKNOWN)
            return Function(
                arguments=[target, self.address.rsplit(".", 1)[1]],
                operator=FunctionType.ATTR_ACCESS,
                arg_count=2,
                output_datatype=output_type,
                output_purpose=Purpose.PROPERTY,
            )
        return self


class UndefinedConcept(ConceptRef):
    pass

    @property
    def reference(self):
        return self

    @property
    def purpose(self):
        return Purpose.UNKNOWN


def address_with_namespace(address: str, namespace: str) -> str:
    existing_ns = address.split(".", 1)[0]
    if "." in address:
        existing_name = address.split(".", 1)[1]
    else:
        existing_name = address
    if existing_name == ALL_ROWS_CONCEPT:
        return address
    if existing_ns == DEFAULT_NAMESPACE:
        return f"{namespace}.{existing_name}"
    return f"{namespace}.{address}"


class Parenthetical(
    DataTyped,
    ConceptArgs,
    Mergeable,
    Namespaced,
    BaseModel,
):
    content: "Expr"

    @field_validator("content", mode="before")
    @classmethod
    def content_validator(cls, v, info: ValidationInfo):
        if isinstance(v, Concept):
            return v.reference
        return v

    def __add__(self, other) -> Union["Parenthetical", "Conditional"]:
        if other is None:
            return self
        elif isinstance(other, (Comparison, Conditional, Parenthetical)):
            return Conditional(left=self, right=other, operator=BooleanOperator.AND)
        raise ValueError(f"Cannot add {self.__class__} and {type(other)}")

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"({str(self.content)})"

    def with_namespace(self, namespace: str) -> Parenthetical:
        return Parenthetical.model_construct(
            content=(
                self.content.with_namespace(namespace)
                if isinstance(self.content, Namespaced)
                else self.content
            )
        )

    def with_merge(self, source: Concept, target: Concept, modifiers: List[Modifier]):
        return Parenthetical.model_construct(
            content=(
                self.content.with_merge(source, target, modifiers)
                if isinstance(self.content, Mergeable)
                else self.content
            )
        )

    def with_reference_replacement(self, source, target):
        return Parenthetical.model_construct(
            content=(
                self.content.with_reference_replacement(source, target)
                if isinstance(self.content, Mergeable)
                else self.content
            )
        )

    @property
    def concept_arguments(self) -> Sequence[ConceptRef]:
        base: List[ConceptRef] = []
        x = self.content
        if isinstance(x, ConceptRef):
            base += [x]
        elif isinstance(x, ConceptArgs):
            base += x.concept_arguments
        return base

    @property
    def row_arguments(self) -> Sequence[ConceptRef]:
        if isinstance(self.content, ConceptArgs):
            return self.content.row_arguments
        return self.concept_arguments

    @property
    def existence_arguments(self) -> Sequence[tuple["ConceptRef", ...]]:
        if isinstance(self.content, ConceptArgs):
            return self.content.existence_arguments
        return []

    @property
    def output_datatype(self):
        return arg_to_datatype(self.content)


class Conditional(Mergeable, ConceptArgs, Namespaced, DataTyped, BaseModel):
    left: Expr
    right: Expr
    operator: BooleanOperator

    @field_validator("left", mode="before")
    @classmethod
    def left_validator(cls, v, info: ValidationInfo):
        if isinstance(v, Concept):
            return v.reference
        return v

    @field_validator("right", mode="before")
    @classmethod
    def right_validator(cls, v, info: ValidationInfo):
        if isinstance(v, Concept):
            return v.reference
        return v

    def __add__(self, other) -> "Conditional":
        if other is None:
            return self
        elif str(other) == str(self):
            return self
        elif isinstance(other, (Comparison, Conditional, Parenthetical)):
            return Conditional.model_construct(
                left=self, right=other, operator=BooleanOperator.AND
            )
        raise ValueError(f"Cannot add {self.__class__} and {type(other)}")

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"{str(self.left)} {self.operator.value} {str(self.right)}"

    def __eq__(self, other):
        if not isinstance(other, Conditional):
            return False
        return (
            self.left == other.left
            and self.right == other.right
            and self.operator == other.operator
        )

    def with_namespace(self, namespace: str) -> "Conditional":
        return Conditional.model_construct(
            left=(
                self.left.with_namespace(namespace)
                if isinstance(self.left, Namespaced)
                else self.left
            ),
            right=(
                self.right.with_namespace(namespace)
                if isinstance(self.right, Namespaced)
                else self.right
            ),
            operator=self.operator,
        )

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> "Conditional":
        return Conditional.model_construct(
            left=(
                self.left.with_merge(source, target, modifiers)
                if isinstance(self.left, Mergeable)
                else self.left
            ),
            right=(
                self.right.with_merge(source, target, modifiers)
                if isinstance(self.right, Mergeable)
                else self.right
            ),
            operator=self.operator,
        )

    def with_reference_replacement(self, source, target):
        return self.__class__.model_construct(
            left=(
                self.left.with_reference_replacement(source, target)
                if isinstance(self.left, Mergeable)
                else self.left
            ),
            right=(
                self.right.with_reference_replacement(source, target)
                if isinstance(self.right, Mergeable)
                else self.right
            ),
            operator=self.operator,
        )

    @property
    def concept_arguments(self) -> Sequence[ConceptRef]:
        output = []
        output += get_concept_arguments(self.left)
        output += get_concept_arguments(self.right)
        return output

    @property
    def row_arguments(self) -> Sequence[ConceptRef]:
        output = []
        output += get_concept_row_arguments(self.left)
        output += get_concept_row_arguments(self.right)
        return output

    @property
    def existence_arguments(self) -> Sequence[tuple[ConceptRef, ...]]:
        output: list[tuple[ConceptRef, ...]] = []
        if isinstance(self.left, ConceptArgs):
            output += self.left.existence_arguments
        if isinstance(self.right, ConceptArgs):
            output += self.right.existence_arguments
        return output

    @property
    def output_datatype(self):
        # a conditional is always a boolean
        return DataType.BOOL

    def decompose(self):
        chunks = []
        if self.operator == BooleanOperator.AND:
            for val in [self.left, self.right]:
                if isinstance(val, Conditional):
                    chunks.extend(val.decompose())
                else:
                    chunks.append(val)
        else:
            chunks.append(self)
        return chunks


class WhereClause(Mergeable, ConceptArgs, Namespaced, BaseModel):
    conditional: Union[SubselectComparison, Comparison, Conditional, Parenthetical]

    def __repr__(self):
        return str(self.conditional)

    def __str__(self):
        return self.__repr__()

    @property
    def concept_arguments(self) -> Sequence[ConceptRef]:
        return self.conditional.concept_arguments

    @property
    def row_arguments(self) -> Sequence[ConceptRef]:
        return self.conditional.row_arguments

    @property
    def existence_arguments(self) -> Sequence[tuple["ConceptRef", ...]]:
        return self.conditional.existence_arguments

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> Self:
        return self.__class__.model_construct(
            conditional=self.conditional.with_merge(source, target, modifiers)
        )

    def with_namespace(self, namespace: str) -> Self:
        return self.__class__.model_construct(
            conditional=self.conditional.with_namespace(namespace)
        )

    def with_reference_replacement(self, source, target):
        return self.__class__.model_construct(
            conditional=self.conditional.with_reference_replacement(source, target)
        )


class HavingClause(WhereClause):
    pass


class Grain(Namespaced, BaseModel):
    components: set[str] = Field(default_factory=set)
    where_clause: Optional["WhereClause"] = None
    _str: str | None = None
    _abstract: bool = False

    def without_condition(self):
        return Grain(components=self.components)

    def with_merge(self, source: Concept, target: Concept, modifiers: List[Modifier]):
        new_components = set()
        for c in self.components:
            if c == source.address:
                new_components.add(target.address)
            else:
                new_components.add(c)
        return Grain.model_construct(components=new_components)

    @classmethod
    def from_concepts(
        cls,
        concepts: Iterable[Concept | ConceptRef | str],
        environment: Environment | None = None,
        where_clause: WhereClause | None = None,
        local_concepts: dict[str, Concept] | None = None,
    ) -> Grain:
        from trilogy.parsing.common import concepts_to_grain_concepts

        x = Grain.model_construct(
            components=concepts_to_grain_concepts(
                concepts, environment=environment, local_concepts=local_concepts
            ),
            where_clause=where_clause,
        )

        return x

    def with_namespace(self, namespace: str) -> "Grain":
        return Grain.model_construct(
            components={address_with_namespace(c, namespace) for c in self.components},
            where_clause=(
                self.where_clause.with_namespace(namespace)
                if self.where_clause
                else None
            ),
        )

    @field_validator("components", mode="before")
    def component_validator(cls, v, info: ValidationInfo):
        output = set()
        if isinstance(v, list):
            for vc in v:
                if isinstance(vc, Addressable):
                    output.add(vc._address)
                else:
                    output.add(vc)
        else:
            output = v
        if not isinstance(output, set):
            raise ValueError(f"Invalid grain component {output}, is not set")
        if not all(isinstance(x, str) for x in output):
            raise ValueError(f"Invalid component {output}")
        return output

    def __add__(self, other: "Grain") -> "Grain":
        if not other:
            return self
        where = self.where_clause
        if other.where_clause:
            if not self.where_clause:
                where = other.where_clause
            elif not other.where_clause == self.where_clause:
                where = WhereClause.model_construct(
                    conditional=Conditional(
                        left=self.where_clause.conditional,
                        right=other.where_clause.conditional,
                        operator=BooleanOperator.AND,
                    )
                )
                # raise NotImplementedError(
                #     f"Cannot merge grains with where clauses, self {self.where_clause} other {other.where_clause}"
                # )
        return Grain(
            components=self.components.union(other.components), where_clause=where
        )

    def __sub__(self, other: "Grain") -> "Grain":
        return Grain.model_construct(
            components=self.components.difference(other.components),
            where_clause=self.where_clause,
        )

    def _gen_abstract(self) -> bool:
        return not self.components or all(
            [c.endswith(ALL_ROWS_CONCEPT) for c in self.components]
        )

    @property
    def abstract(self):
        if not self._abstract:
            self._abstract = self._gen_abstract()
        return self._abstract

    def __eq__(self, other: object):
        if isinstance(other, list):
            if all([isinstance(c, Concept) for c in other]):
                return self.components == set([c.address for c in other])
            return False
        if not isinstance(other, Grain):
            return False
        if self.components == other.components:
            return True
        return False

    def issubset(self, other: "Grain"):
        return self.components.issubset(other.components)

    def union(self, other: "Grain"):
        addresses = self.components.union(other.components)
        return Grain(components=addresses, where_clause=self.where_clause)

    def isdisjoint(self, other: "Grain"):
        return self.components.isdisjoint(other.components)

    def intersection(self, other: "Grain") -> "Grain":
        intersection = self.components.intersection(other.components)
        return Grain(components=intersection)

    def _gen_str(self) -> str:
        if self.abstract:
            base = "Grain<Abstract>"
        else:
            base = "Grain<" + ",".join(sorted(self.components)) + ">"
        if self.where_clause:
            base += f"|{str(self.where_clause)}"
        return base

    def __str__(self):
        if not self._str:
            self._str = self._gen_str()
        return self._str

    def __radd__(self, other) -> "Grain":
        if other == 0:
            return self
        else:
            return self.__add__(other)


class Comparison(ConceptArgs, Mergeable, DataTyped, Namespaced, BaseModel):
    left: Union[
        int,
        str,
        float,
        list,
        bool,
        datetime,
        date,
        Function,
        ConceptRef,
        Conditional,
        DataType,
        Comparison,
        FunctionCallWrapper,
        Parenthetical,
        MagicConstants,
        WindowItem,
        AggregateWrapper,
        FilterItem,
    ]
    right: Union[
        int,
        str,
        float,
        list,
        bool,
        date,
        datetime,
        ConceptRef,
        Function,
        Conditional,
        DataType,
        Comparison,
        FunctionCallWrapper,
        Parenthetical,
        MagicConstants,
        WindowItem,
        AggregateWrapper,
        TupleWrapper,
        FilterItem,
    ]
    operator: ComparisonOperator

    @field_validator("left", mode="before")
    @classmethod
    def left_validator(cls, v, info: ValidationInfo):
        if isinstance(v, Concept):
            return v.reference
        return v

    @field_validator("right", mode="before")
    @classmethod
    def right_validator(cls, v, info: ValidationInfo):
        if isinstance(v, Concept):
            return v.reference
        return v

    @model_validator(mode="after")
    def validate_comparison(self):
        left_type = arg_to_datatype(self.left)
        right_type = arg_to_datatype(self.right)
        left_name = (
            left_type.name if isinstance(left_type, DataType) else str(left_type)
        )
        right_name = (
            right_type.name if isinstance(right_type, DataType) else str(right_type)
        )
        if self.operator in (ComparisonOperator.IS, ComparisonOperator.IS_NOT):
            if self.right != MagicConstants.NULL and DataType.BOOL != right_type:
                raise SyntaxError(
                    f"Cannot use {self.operator.value} with non-null or boolean value {self.right}"
                )
        elif self.operator in (ComparisonOperator.IN, ComparisonOperator.NOT_IN):

            if isinstance(right_type, ArrayType) and not is_compatible_datatype(
                left_type, right_type.value_data_type
            ):
                raise SyntaxError(
                    f"Cannot compare {left_type} and {right_type} with operator {self.operator} in {str(self)}"
                )
            elif isinstance(self.right, Concept) and not is_compatible_datatype(
                left_type, right_type
            ):
                raise SyntaxError(
                    f"Cannot compare {left_name} and {right_name} with operator {self.operator} in {str(self)}"
                )
        else:
            if not is_compatible_datatype(left_type, right_type):
                raise SyntaxError(
                    f"Cannot compare {left_name} ({self.left}) and {right_name} ({self.right}) of different types with operator {self.operator.value} in {str(self)}"
                )

        return self

    def __add__(self, other):
        if other is None:
            return self
        if not isinstance(other, (Comparison, Conditional, Parenthetical)):
            raise ValueError("Cannot add Comparison to non-Comparison")
        if other == self:
            return self
        return Conditional(left=self, right=other, operator=BooleanOperator.AND)

    def __repr__(self):
        if isinstance(self.left, Concept):
            left = self.left.address
        else:
            left = str(self.left)
        if isinstance(self.right, Concept):
            right = self.right.address
        else:
            right = str(self.right)
        return f"{left} {self.operator.value} {right}"

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if not isinstance(other, Comparison):
            return False
        return (
            self.left == other.left
            and self.right == other.right
            and self.operator == other.operator
        )

    def with_merge(self, source: Concept, target: Concept, modifiers: List[Modifier]):
        return self.__class__.model_construct(
            left=(
                self.left.with_merge(source, target, modifiers)
                if isinstance(self.left, Mergeable)
                else self.left
            ),
            right=(
                self.right.with_merge(source, target, modifiers)
                if isinstance(self.right, Mergeable)
                else self.right
            ),
            operator=self.operator,
        )

    def with_reference_replacement(self, source, target):
        return self.__class__.model_construct(
            left=(
                self.left.with_reference_replacement(source, target)
                if isinstance(self.left, Mergeable)
                else self.left
            ),
            right=(
                self.right.with_reference_replacement(source, target)
                if isinstance(self.right, Mergeable)
                else self.right
            ),
            operator=self.operator,
        )

    def with_namespace(self, namespace: str):
        return self.__class__.model_construct(
            left=(
                self.left.with_namespace(namespace)
                if isinstance(self.left, Namespaced)
                else self.left
            ),
            right=(
                self.right.with_namespace(namespace)
                if isinstance(self.right, Namespaced)
                else self.right
            ),
            operator=self.operator,
        )

    @property
    def concept_arguments(self) -> List[ConceptRef]:
        """Return concepts directly referenced in where clause"""
        output = []
        output += get_concept_arguments(self.left)
        output += get_concept_arguments(self.right)
        return output

    @property
    def row_arguments(self) -> List[ConceptRef]:
        output = []
        output += get_concept_row_arguments(self.left)
        output += get_concept_row_arguments(self.right)
        return output

    @property
    def existence_arguments(self) -> List[Tuple[ConceptRef, ...]]:
        """Return concepts directly referenced in where clause"""
        output: List[Tuple[ConceptRef, ...]] = []
        if isinstance(self.left, ConceptArgs):
            output += self.left.existence_arguments
        if isinstance(self.right, ConceptArgs):
            output += self.right.existence_arguments
        return output

    @property
    def output_datatype(self):
        # a conditional is always a boolean
        return DataType.BOOL


class SubselectComparison(Comparison):
    def __eq__(self, other):
        if not isinstance(other, SubselectComparison):
            return False

        comp = (
            self.left == other.left
            and self.right == other.right
            and self.operator == other.operator
        )
        return comp

    @property
    def row_arguments(self) -> List[ConceptRef]:
        return get_concept_row_arguments(self.left)

    @property
    def existence_arguments(self) -> list[tuple["ConceptRef", ...]]:
        return [tuple(get_concept_arguments(self.right))]


class Concept(Addressable, DataTyped, ConceptArgs, Mergeable, Namespaced, BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    name: str
    datatype: DataType | TraitDataType | ArrayType | StructType | MapType | NumericType
    purpose: Purpose
    derivation: Derivation = Derivation.ROOT
    granularity: Granularity = Granularity.MULTI_ROW
    metadata: Metadata = Field(
        default_factory=lambda: Metadata(description=None, line_number=None),
        validate_default=True,
    )
    lineage: Optional[
        Union[
            Function,
            WindowItem,
            FilterItem,
            AggregateWrapper,
            RowsetItem,
            MultiSelectLineage,
            Comparison,
        ]
    ] = None
    namespace: str = Field(default=DEFAULT_NAMESPACE, validate_default=True)
    keys: Optional[set[str]] = None
    grain: "Grain" = Field(default=None, validate_default=True)  # type: ignore
    modifiers: List[Modifier] = Field(default_factory=list)  # type: ignore
    pseudonyms: set[str] = Field(default_factory=set)

    def duplicate(self) -> Concept:
        return self.model_copy(deep=True)

    def __hash__(self):
        return hash(
            f"{self.name}+{self.datatype}+ {self.purpose} + {str(self.lineage)} + {self.namespace} + {str(self.grain)} + {str(self.keys)}"
        )

    def __repr__(self):
        base = f"{self.address}@{self.grain}"
        return base

    @property
    def is_internal(self) -> bool:
        return self.namespace.startswith("_") or self.name.startswith("_")

    @property
    def reference(self) -> ConceptRef:
        return ConceptRef.model_construct(
            address=self.address,
            datatype=self.output_datatype,
            metadata=self.metadata,
        )

    @property
    def output_datatype(self):
        return self.datatype

    @classmethod
    def calculate_is_aggregate(cls, lineage):
        if lineage and isinstance(lineage, Function):
            if lineage.operator in FunctionClass.AGGREGATE_FUNCTIONS.value:
                return True
        if (
            lineage
            and isinstance(lineage, AggregateWrapper)
            and lineage.function.operator in FunctionClass.AGGREGATE_FUNCTIONS.value
        ):
            return True
        return False

    @property
    def is_aggregate(self):
        base = getattr(self, "_is_aggregate", None)
        if base:
            return base
        setattr(self, "_is_aggregate", self.calculate_is_aggregate(self.lineage))
        return self._is_aggregate

    def with_merge(self, source: Self, target: Self, modifiers: List[Modifier]) -> Self:
        if self.address == source.address:
            new = target.with_grain(self.grain.with_merge(source, target, modifiers))
            new.pseudonyms.add(self.address)
            return new
        if not self.grain.components and not self.lineage and not self.keys:
            return self
        return self.__class__.model_construct(
            name=self.name,
            datatype=self.datatype,
            purpose=self.purpose,
            metadata=self.metadata,
            derivation=self.derivation,
            granularity=self.granularity,
            lineage=(
                self.lineage.with_merge(source, target, modifiers)
                if self.lineage
                else None
            ),
            grain=self.grain.with_merge(source, target, modifiers),
            namespace=self.namespace,
            keys=(
                set(x if x != source.address else target.address for x in self.keys)
                if self.keys
                else None
            ),
            modifiers=self.modifiers,
            pseudonyms=self.pseudonyms,
        )

    @field_validator("namespace", mode="plain")
    @classmethod
    def namespace_validation(cls, v):
        return v or DEFAULT_NAMESPACE

    @field_validator("metadata", mode="before")
    @classmethod
    def metadata_validation(cls, v):
        v = v or Metadata()
        return v

    @field_validator("purpose", mode="after")
    @classmethod
    def purpose_validation(cls, v):
        if v == Purpose.AUTO:
            raise ValueError("Cannot set purpose to AUTO")
        return v

    @field_validator("grain", mode="before")
    @classmethod
    def parse_grain(cls, v, info: ValidationInfo) -> Grain:

        # this is silly - rethink how we do grains
        values = info.data

        if not v and values.get("purpose", None) == Purpose.KEY:
            v = Grain(
                components={
                    f'{values.get("namespace", DEFAULT_NAMESPACE)}.{values["name"]}'
                }
            )
        elif not v and values.get("purpose", None) == Purpose.PROPERTY:
            v = Grain(components=values.get("keys", set()) or set())
        elif (
            "lineage" in values
            and isinstance(values["lineage"], AggregateWrapper)
            and values["lineage"].by
        ):
            v = Grain(components={c.address for c in values["lineage"].by})
        elif not v:
            v = Grain(components=set())
        elif isinstance(v, Grain):
            pass
        elif isinstance(v, Concept):
            v = Grain(components={v.address})
        elif isinstance(v, dict):
            v = Grain.model_validate(v)
        else:
            raise SyntaxError(f"Invalid grain {v} for concept {values['name']}")
        return v

    def __eq__(self, other: object):
        if isinstance(other, str):
            if self.address == other:
                return True
        if isinstance(other, ConceptRef):
            return self.address == other.address
        if not isinstance(other, Concept):
            return False
        return (
            self.name == other.name
            and self.datatype == other.datatype
            and self.purpose == other.purpose
            and self.namespace == other.namespace
            and self.grain == other.grain
            and self.derivation == other.derivation
            and self.granularity == other.granularity
            # and self.keys == other.keys
        )

    def __str__(self):
        grain = str(self.grain) if self.grain else "Grain<>"
        return f"{self.namespace}.{self.name}@{grain}"

    @property
    def address(self) -> str:
        return f"{self.namespace}.{self.name}"

    @property
    def output(self) -> "Concept":
        return self

    @property
    def safe_address(self) -> str:
        if self.namespace == DEFAULT_NAMESPACE:
            return self.name.replace(".", "_")
        elif self.namespace:
            return f"{self.namespace.replace('.','_')}_{self.name.replace('.','_')}"
        return self.name.replace(".", "_")

    def with_namespace(self, namespace: str) -> Self:
        return self.__class__.model_construct(
            name=self.name,
            datatype=self.datatype,
            purpose=self.purpose,
            granularity=self.granularity,
            derivation=self.derivation,
            metadata=self.metadata,
            lineage=self.lineage.with_namespace(namespace) if self.lineage else None,
            grain=(
                self.grain.with_namespace(namespace)
                if self.grain
                else Grain(components=set())
            ),
            namespace=(
                namespace + "." + self.namespace
                if self.namespace != DEFAULT_NAMESPACE
                else namespace
            ),
            keys=(
                set([address_with_namespace(x, namespace) for x in self.keys])
                if self.keys
                else None
            ),
            modifiers=self.modifiers,
            pseudonyms={address_with_namespace(v, namespace) for v in self.pseudonyms},
        )

    def get_select_grain_and_keys(
        self, grain: Grain, environment: Environment
    ) -> Tuple[
        Function
        | WindowItem
        | FilterItem
        | AggregateWrapper
        | RowsetItem
        | MultiSelectLineage
        | Comparison
        | None,
        Grain,
        set[str] | None,
    ]:
        new_lineage = self.lineage
        final_grain = grain if not self.grain.components else self.grain
        keys = self.keys

        if self.is_aggregate and grain.components and isinstance(new_lineage, Function):
            grain_components: list[ConceptRef | Concept] = [
                environment.concepts[c].reference for c in grain.components
            ]
            new_lineage = AggregateWrapper.model_construct(
                function=new_lineage, by=grain_components
            )
            final_grain = grain
            keys = set(grain.components)
        elif (
            grain
            and new_lineage
            and isinstance(new_lineage, AggregateWrapper)
            and not new_lineage.by
        ):
            grain_components = [
                environment.concepts[c].reference for c in grain.components
            ]
            new_lineage = AggregateWrapper.model_construct(
                function=new_lineage.function, by=grain_components
            )
            final_grain = grain
            keys = set([x.address for x in new_lineage.by])
        elif self.derivation == Derivation.BASIC:

            pkeys: set[str] = set()
            assert new_lineage
            for x_ref in new_lineage.concept_arguments:
                x = environment.concepts[x_ref.address]
                if isinstance(x, UndefinedConcept):
                    continue
                _, _, parent_keys = x.get_select_grain_and_keys(grain, environment)
                if parent_keys:
                    pkeys.update(parent_keys)
            raw_keys = pkeys
            # deduplicate

            final_grain = Grain.from_concepts(raw_keys, environment)
            keys = final_grain.components
        return new_lineage, final_grain, keys

    def set_select_grain(self, grain: Grain, environment: Environment) -> Self:
        """Assign a mutable concept the appropriate grain/keys for a select"""
        new_lineage, final_grain, keys = self.get_select_grain_and_keys(
            grain, environment
        )
        return self.__class__.model_construct(
            name=self.name,
            datatype=self.datatype,
            purpose=self.purpose,
            granularity=self.granularity,
            derivation=self.derivation,
            metadata=self.metadata,
            lineage=new_lineage,
            grain=final_grain,
            namespace=self.namespace,
            keys=keys,
            modifiers=self.modifiers,
            pseudonyms=self.pseudonyms,
        )

    def with_grain(self, grain: Optional["Grain"] = None) -> Self:

        return self.__class__.model_construct(
            name=self.name,
            datatype=self.datatype,
            purpose=self.purpose,
            metadata=self.metadata,
            granularity=self.granularity,
            derivation=self.derivation,
            lineage=self.lineage,
            grain=grain if grain else Grain.model_construct(components=set()),
            namespace=self.namespace,
            keys=self.keys,
            modifiers=self.modifiers,
            pseudonyms=self.pseudonyms,
        )

    @cached_property
    def sources(self) -> List["ConceptRef"]:
        if self.lineage:
            output: List[ConceptRef] = []

            def get_sources(
                expr: Union[
                    Function,
                    WindowItem,
                    FilterItem,
                    AggregateWrapper,
                    RowsetItem,
                    MultiSelectLineage,
                    Comparison,
                ],
                output: List[ConceptRef],
            ):

                for item in expr.concept_arguments:
                    if isinstance(item, (ConceptRef,)):
                        if item.address == self.address:
                            raise SyntaxError(
                                f"Concept {self.address} references itself"
                            )
                        output.append(item)

                        # output += item.sources

            get_sources(self.lineage, output)
            return output
        return []

    @property
    def concept_arguments(self) -> List[ConceptRef]:
        return self.lineage.concept_arguments if self.lineage else []

    @classmethod
    def calculate_derivation(self, lineage, purpose: Purpose) -> Derivation:
        from trilogy.core.models.build import (
            BuildAggregateWrapper,
            BuildComparison,
            BuildFilterItem,
            BuildFunction,
            BuildMultiSelectLineage,
            BuildRowsetItem,
            BuildWindowItem,
        )

        if lineage and isinstance(lineage, (BuildWindowItem, WindowItem)):
            return Derivation.WINDOW
        elif lineage and isinstance(lineage, (BuildFilterItem, FilterItem)):
            return Derivation.FILTER
        elif lineage and isinstance(lineage, (BuildAggregateWrapper, AggregateWrapper)):
            return Derivation.AGGREGATE
        # elif lineage and isinstance(lineage, (BuildParenthetical, Parenthetical)):
        #     return Derivation.PARENTHETICAL
        elif lineage and isinstance(lineage, (BuildRowsetItem, RowsetItem)):
            return Derivation.ROWSET
        elif lineage and isinstance(lineage, BuildComparison):
            return Derivation.BASIC
        elif lineage and isinstance(
            lineage, (BuildMultiSelectLineage, MultiSelectLineage)
        ):
            return Derivation.MULTISELECT
        elif (
            lineage
            and isinstance(lineage, (BuildFunction, Function))
            and lineage.operator in FunctionClass.AGGREGATE_FUNCTIONS.value
        ):
            return Derivation.AGGREGATE
        elif (
            lineage
            and isinstance(lineage, (BuildFunction, Function))
            and lineage.operator in FunctionClass.ONE_TO_MANY.value
        ):
            return Derivation.UNNEST
        elif (
            lineage
            and isinstance(lineage, (BuildFunction, Function))
            and lineage.operator == FunctionType.RECURSE_EDGE
        ):
            return Derivation.RECURSIVE
        elif (
            lineage
            and isinstance(lineage, (BuildFunction, Function))
            and lineage.operator == FunctionType.UNION
        ):
            return Derivation.UNION
        elif (
            lineage
            and isinstance(lineage, (BuildFunction, Function))
            and lineage.operator == FunctionType.GROUP
        ):
            return Derivation.GROUP_TO
        elif (
            lineage
            and isinstance(lineage, (BuildFunction, Function))
            and lineage.operator in FunctionClass.SINGLE_ROW.value
        ):
            return Derivation.CONSTANT

        elif lineage and isinstance(lineage, (BuildFunction, Function)):
            if not lineage.concept_arguments:
                return Derivation.CONSTANT
            elif all(
                [x.derivation == Derivation.CONSTANT for x in lineage.concept_arguments]
            ):
                return Derivation.CONSTANT
            return Derivation.BASIC
        elif purpose == Purpose.CONSTANT:
            return Derivation.CONSTANT
        return Derivation.ROOT

    @classmethod
    def calculate_granularity(cls, derivation: Derivation, grain: Grain, lineage):
        from trilogy.core.models.build import BuildFunction

        if derivation == Derivation.CONSTANT:
            return Granularity.SINGLE_ROW
        elif derivation == Derivation.AGGREGATE:
            if all([x.endswith(ALL_ROWS_CONCEPT) for x in grain.components]):
                return Granularity.SINGLE_ROW
        elif (
            lineage
            and isinstance(lineage, (Function, BuildFunction))
            and lineage.operator
            in (FunctionType.UNNEST, FunctionType.UNION, FunctionType.DATE_SPINE)
        ):
            return Granularity.MULTI_ROW
        elif lineage and all(
            [x.granularity == Granularity.SINGLE_ROW for x in lineage.concept_arguments]
        ):
            return Granularity.SINGLE_ROW
        return Granularity.MULTI_ROW

    # @property
    # def granularity(self) -> Granularity:
    #     return self.calculate_granularity(self.derivation, self.grain, self.lineage)

    def with_filter(
        self,
        condition: Conditional | Comparison | Parenthetical,
        environment: Environment | None = None,
    ) -> "Concept":
        from trilogy.utility import string_to_hash

        if self.lineage and isinstance(self.lineage, FilterItem):
            if self.lineage.where.conditional == condition:
                return self
        hash = string_to_hash(self.name + str(condition))
        new_lineage = FilterItem(
            content=self.reference, where=WhereClause(conditional=condition)
        )
        new = Concept.model_construct(
            name=f"{self.name}_filter_{hash}",
            datatype=self.datatype,
            purpose=self.purpose,
            derivation=self.calculate_derivation(new_lineage, self.purpose),
            granularity=self.granularity,
            metadata=self.metadata,
            lineage=new_lineage,
            keys=(self.keys if self.purpose == Purpose.PROPERTY else None),
            grain=self.grain if self.grain else Grain(components=set()),
            namespace=self.namespace,
            modifiers=self.modifiers,
            pseudonyms=self.pseudonyms,
        )
        if environment:
            environment.add_concept(new)
        return new


class UndefinedConceptFull(Concept, Mergeable, Namespaced):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str
    line_no: int | None = None
    datatype: (
        DataType | TraitDataType | ArrayType | StructType | MapType | NumericType
    ) = DataType.UNKNOWN
    purpose: Purpose = Purpose.UNKNOWN

    @property
    def reference(self) -> UndefinedConcept:
        return UndefinedConcept(address=self.address)


class OrderItem(Mergeable, ConceptArgs, Namespaced, BaseModel):
    # this needs to be a full concept as it may not exist in environment
    expr: Expr
    order: Ordering

    @field_validator("expr", mode="before")
    def enforce_reference(cls, v):
        if isinstance(v, Concept):
            return v.reference
        return v

    def with_namespace(self, namespace: str) -> "OrderItem":
        return OrderItem.model_construct(
            expr=(
                self.expr.with_namespace(namespace)
                if isinstance(self.expr, Namespaced)
                else self.expr
            ),
            order=self.order,
        )

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> "OrderItem":
        return OrderItem.model_construct(
            expr=(
                self.expr.with_merge(source, target, modifiers)
                if isinstance(self.expr, Mergeable)
                else self.expr
            ),
            order=self.order,
        )

    def with_reference_replacement(self, source, target):
        return OrderItem.model_construct(
            expr=(
                self.expr.with_reference_replacement(source, target)
                if isinstance(self.expr, Mergeable)
                else self.expr
            ),
            order=self.order,
        )

    @property
    def concept_arguments(self) -> Sequence[ConceptRef]:
        return get_concept_arguments(self.expr)

    @property
    def row_arguments(self) -> Sequence[ConceptRef]:
        if isinstance(self.expr, ConceptArgs):
            return self.expr.row_arguments
        return self.concept_arguments

    @property
    def existence_arguments(self) -> Sequence[tuple["ConceptRef", ...]]:
        if isinstance(self.expr, ConceptArgs):
            return self.expr.existence_arguments
        return []

    @property
    def output_datatype(self):
        return arg_to_datatype(self.expr)


class WindowItem(DataTyped, ConceptArgs, Mergeable, Namespaced, BaseModel):
    type: WindowType
    content: FuncArgs
    order_by: List["OrderItem"]
    over: List["ConceptRef"] = Field(default_factory=list)
    index: Optional[int] = None

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"{self.type.value} {self.content} by {self.index} over {self.over} order {self.order_by}"

    @field_validator("content", mode="before")
    def enforce_concept_ref(cls, v):
        if isinstance(v, Concept):
            return ConceptRef(address=v.address, datatype=v.datatype)
        return v

    @field_validator("over", mode="before")
    def enforce_concept_ref_over(cls, v):
        final = []
        for item in v:
            if isinstance(item, Concept):
                final.append(ConceptRef(address=item.address, datatype=item.datatype))
            else:
                final.append(item)
        return final

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> "WindowItem":
        output = WindowItem.model_construct(
            type=self.type,
            content=(
                self.content.with_merge(source, target, modifiers)
                if isinstance(self.content, Mergeable)
                else self.content
            ),
            over=[x.with_merge(source, target, modifiers) for x in self.over],
            order_by=[x.with_merge(source, target, modifiers) for x in self.order_by],
            index=self.index,
        )
        return output

    def with_reference_replacement(self, source, target):
        return WindowItem.model_construct(
            type=self.type,
            content=self.content.with_reference_replacement(source, target),
            over=[x.with_reference_replacement(source, target) for x in self.over],
            order_by=[
                x.with_reference_replacement(source, target) for x in self.order_by
            ],
            index=self.index,
        )

    def with_namespace(self, namespace: str) -> "WindowItem":
        return WindowItem.model_construct(
            type=self.type,
            content=(
                self.content.with_namespace(namespace)
                if isinstance(self.content, Namespaced)
                else self.content
            ),
            over=[x.with_namespace(namespace) for x in self.over],
            order_by=[x.with_namespace(namespace) for x in self.order_by],
            index=self.index,
        )

    @property
    def concept_arguments(self) -> List[ConceptRef]:
        output = []
        output += get_concept_arguments(self.content)
        for order in self.order_by:
            output += get_concept_arguments(order)
        for item in self.over:
            output += get_concept_arguments(item)
        return output

    @property
    def output_datatype(self):
        if self.type in (WindowType.RANK, WindowType.ROW_NUMBER):
            return DataType.INTEGER
        return self.content.output_datatype


def get_basic_type(
    type: DataType | ArrayType | StructType | MapType | NumericType | TraitDataType,
) -> DataType:
    if isinstance(type, ArrayType):
        return DataType.ARRAY
    if isinstance(type, StructType):
        return DataType.STRUCT
    if isinstance(type, MapType):
        return DataType.MAP
    if isinstance(type, NumericType):
        return DataType.NUMERIC
    if isinstance(type, TraitDataType):
        return get_basic_type(type.type)
    return type


class CaseWhen(Namespaced, DataTyped, ConceptArgs, Mergeable, BaseModel):
    comparison: Conditional | SubselectComparison | Comparison
    expr: "Expr"

    @field_validator("expr", mode="before")
    def enforce_reference(cls, v):
        if isinstance(v, Concept):
            return v.reference
        return v

    @property
    def output_datatype(self):
        return arg_to_datatype(self.expr)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"WHEN {str(self.comparison)} THEN {str(self.expr)}"

    @property
    def concept_arguments(self):
        return get_concept_arguments(self.comparison) + get_concept_arguments(self.expr)

    @property
    def concept_row_arguments(self):
        return get_concept_row_arguments(self.comparison) + get_concept_row_arguments(
            self.expr
        )

    def with_namespace(self, namespace: str) -> CaseWhen:
        return CaseWhen.model_construct(
            comparison=self.comparison.with_namespace(namespace),
            expr=(
                self.expr.with_namespace(namespace)
                if isinstance(
                    self.expr,
                    Namespaced,
                )
                else self.expr
            ),
        )

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> CaseWhen:
        return CaseWhen.model_construct(
            comparison=self.comparison.with_merge(source, target, modifiers),
            expr=(
                self.expr.with_merge(source, target, modifiers)
                if isinstance(self.expr, Mergeable)
                else self.expr
            ),
        )

    def with_reference_replacement(self, source, target):
        return CaseWhen.model_construct(
            comparison=self.comparison.with_reference_replacement(source, target),
            expr=(
                self.expr.with_reference_replacement(source, target)
                if isinstance(self.expr, Mergeable)
                else self.expr
            ),
        )


class CaseElse(Namespaced, ConceptArgs, DataTyped, Mergeable, BaseModel):
    expr: "Expr"
    # this ensures that it's easily differentiable from CaseWhen
    discriminant: ComparisonOperator = ComparisonOperator.ELSE

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"ELSE {str(self.expr)}"

    @property
    def output_datatype(self):
        return arg_to_datatype(self.expr)

    @field_validator("expr", mode="before")
    def enforce_expr(cls, v):
        if isinstance(v, Concept):
            return ConceptRef(address=v.address, datatype=v.datatype)
        return v

    @property
    def concept_arguments(self):
        return get_concept_arguments(self.expr)

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> CaseElse:
        return CaseElse.model_construct(
            discriminant=self.discriminant,
            expr=(
                self.expr.with_merge(source, target, modifiers)
                if isinstance(self.expr, Mergeable)
                else self.expr
            ),
        )

    def with_reference_replacement(self, source, target):
        return CaseElse.model_construct(
            discriminant=self.discriminant,
            expr=(
                self.expr.with_reference_replacement(
                    source,
                    target,
                )
                if isinstance(self.expr, Mergeable)
                else self.expr
            ),
        )

    def with_namespace(self, namespace: str) -> CaseElse:
        return CaseElse.model_construct(
            discriminant=self.discriminant,
            expr=(
                self.expr.with_namespace(namespace)
                if isinstance(
                    self.expr,
                    Namespaced,
                )
                else self.expr
            ),
        )


def get_concept_row_arguments(expr) -> List["ConceptRef"]:
    output = []
    if isinstance(expr, ConceptRef):
        output += [expr]

    elif isinstance(expr, ConceptArgs):
        output += expr.row_arguments
    return output


def get_concept_arguments(expr) -> List["ConceptRef"]:
    output = []
    if isinstance(expr, ConceptRef):
        output += [expr]

    elif isinstance(
        expr,
        ConceptArgs,
    ):
        output += expr.concept_arguments
    return output


def args_to_pretty(input: set[DataType]) -> str:
    return ", ".join(sorted([f"'{x.value}'" for x in input if x != DataType.UNKNOWN]))


class Function(DataTyped, ConceptArgs, Mergeable, Namespaced, BaseModel):
    operator: FunctionType
    arg_count: int = Field(default=1)
    output_datatype: (
        DataType | ArrayType | StructType | MapType | NumericType | TraitDataType
    )
    output_purpose: Purpose
    valid_inputs: Optional[
        Union[
            Set[DataType],
            List[Set[DataType]],
        ]
    ] = None
    arguments: Sequence[FuncArgs]

    class Config:
        frozen = True

    def __repr__(self):
        return f'{self.operator.value}({",".join([str(a) for a in self.arguments])})'

    def __str__(self):
        return self.__repr__()

    @property
    def datatype(self):
        return self.output_datatype

    @field_validator("arguments", mode="before")
    @classmethod
    def parse_arguments(cls, v, info: ValidationInfo):
        from trilogy.core.models.build import BuildConcept
        from trilogy.parsing.exceptions import ParseError

        values = info.data
        arg_count = len(v)
        final = []
        for x in v:
            if isinstance(x, Concept) and not isinstance(x, BuildConcept):
                final.append(x.reference)
            else:
                final.append(x)
        v = final
        target_arg_count = values["arg_count"]
        operator_name = values["operator"].name
        # surface right error
        if "valid_inputs" not in values:
            return v
        valid_inputs = values["valid_inputs"]
        if not arg_count <= target_arg_count:
            if target_arg_count != InfiniteFunctionArgs:
                raise ParseError(
                    f"Incorrect argument count to {operator_name} function, expects"
                    f" {target_arg_count}, got {arg_count}"
                )
        # if all arguments can be any of the set type
        # turn this into an array for validation
        if isinstance(valid_inputs, set):
            valid_inputs = [valid_inputs for _ in v]
        elif not valid_inputs:
            return v
        for idx, arg in enumerate(v):
            if (
                isinstance(arg, ConceptRef)
                and get_basic_type(arg.datatype.data_type) not in valid_inputs[idx]
            ):
                if arg.datatype != DataType.UNKNOWN:

                    raise TypeError(
                        f"Invalid argument type '{arg.datatype.data_type.value}' passed into {operator_name} function in position {idx+1}"
                        f" from concept: {arg.name}. Valid: {args_to_pretty(valid_inputs[idx])}."
                    )
            if (
                isinstance(arg, Function)
                and get_basic_type(arg.output_datatype) not in valid_inputs[idx]
            ):
                if arg.output_datatype != DataType.UNKNOWN:
                    raise TypeError(
                        f"Invalid argument type {arg.output_datatype}' passed into"
                        f" {operator_name} function from function {arg.operator.name} in position {idx+1}. Valid: {args_to_pretty(valid_inputs[idx])}"
                    )
            # check constants
            comparisons: List[Tuple[Type, DataType]] = [
                (str, DataType.STRING),
                (int, DataType.INTEGER),
                (float, DataType.FLOAT),
                (bool, DataType.BOOL),
                (DatePart, DataType.DATE_PART),
            ]
            for ptype, dtype in comparisons:
                if (
                    isinstance(arg, ptype)
                    and get_basic_type(dtype) in valid_inputs[idx]
                ):
                    # attempt to exit early to avoid checking all types
                    break
                elif isinstance(arg, ptype):
                    if isinstance(arg, str) and DataType.DATE_PART in valid_inputs[idx]:
                        if arg not in [x.value for x in DatePart]:
                            pass
                        else:
                            break
                    raise TypeError(
                        f'Invalid {dtype} constant passed into {operator_name} "{arg}", expecting one of {valid_inputs[idx]}'
                    )
        return v

    def with_reference_replacement(self, source: str, target: Expr | ArgBinding):
        from trilogy.core.functions import arg_to_datatype, merge_datatypes

        nargs = [
            (
                c.with_reference_replacement(
                    source,
                    target,
                )
                if isinstance(
                    c,
                    Mergeable,
                )
                else c
            )
            for c in self.arguments
        ]
        if self.output_datatype == DataType.UNKNOWN:
            new_output = merge_datatypes([arg_to_datatype(x) for x in nargs])

            if self.operator == FunctionType.ATTR_ACCESS:
                if isinstance(new_output, StructType):
                    new_output = new_output.field_types[str(nargs[1])]
        else:
            new_output = self.output_datatype
        # this is not ideal - see hacky logic for datatypes above
        # we need to figure out how to patch properly
        # should use function factory, but does not have environment access
        # probably move all datatype resolution to build?
        return Function.model_construct(
            operator=self.operator,
            arguments=nargs,
            output_datatype=new_output,
            output_purpose=self.output_purpose,
            valid_inputs=self.valid_inputs,
            arg_count=self.arg_count,
        )

    def with_namespace(self, namespace: str) -> "Function":
        return Function.model_construct(
            operator=self.operator,
            arguments=[
                (
                    c.with_namespace(namespace)
                    if isinstance(
                        c,
                        Namespaced,
                    )
                    else c
                )
                for c in self.arguments
            ],
            output_datatype=self.output_datatype,
            output_purpose=self.output_purpose,
            valid_inputs=self.valid_inputs,
            arg_count=self.arg_count,
        )

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> "Function":
        return Function.model_construct(
            operator=self.operator,
            arguments=[
                (
                    c.with_merge(source, target, modifiers)
                    if isinstance(
                        c,
                        Mergeable,
                    )
                    else c
                )
                for c in self.arguments
            ],
            output_datatype=self.output_datatype,
            output_purpose=self.output_purpose,
            valid_inputs=self.valid_inputs,
            arg_count=self.arg_count,
        )

    @property
    def concept_arguments(self) -> List[ConceptRef]:
        base = []
        for arg in self.arguments:
            base += get_concept_arguments(arg)
        return base


class FunctionCallWrapper(
    DataTyped,
    ConceptArgs,
    Mergeable,
    Namespaced,
    BaseModel,
):
    content: Expr
    name: str
    args: List[Expr]

    def __str__(self):
        return f'@{self.name}({",".join([str(x) for x in self.args])})'

    def with_namespace(self, namespace) -> "FunctionCallWrapper":
        return FunctionCallWrapper.model_construct(
            content=(
                self.content.with_namespace(namespace)
                if isinstance(self.content, Namespaced)
                else self.content
            ),
            name=self.name,
            args=[
                x.with_namespace(namespace) if isinstance(x, Namespaced) else x
                for x in self.args
            ],
        )

    def with_reference_replacement(self, source, target):
        raise NotImplementedError("Cannot reference replace")
        return self

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> "FunctionCallWrapper":
        return FunctionCallWrapper.model_construct(
            content=(
                self.content.with_merge(source, target, modifiers)
                if isinstance(self.content, Mergeable)
                else self.content
            ),
            name=self.name,
            args=[
                (
                    x.with_merge(source, target, modifiers)
                    if isinstance(x, Mergeable)
                    else x
                )
                for x in self.args
            ],
        )

    @property
    def concept_arguments(self) -> Sequence[ConceptRef]:
        base: List[ConceptRef] = []
        x = self.content
        if isinstance(x, ConceptRef):
            base += [x]
        elif isinstance(x, ConceptArgs):
            base += x.concept_arguments
        return base

    @property
    def output_datatype(self):
        return arg_to_datatype(self.content)


class AggregateWrapper(Mergeable, DataTyped, ConceptArgs, Namespaced, BaseModel):
    function: Function
    by: List[ConceptRef | Concept] = Field(default_factory=list)

    @field_validator("by", mode="before")
    @classmethod
    def enforce_concept_ref(cls, v):
        output = []
        for item in v:
            if isinstance(item, Concept):
                output.append(item.reference)
            else:
                output.append(item)
        return output

    def __str__(self):
        grain_str = [str(c) for c in self.by] if self.by else "abstract"
        return f"{str(self.function)}<{grain_str}>"

    @property
    def datatype(self):
        return self.function.datatype

    @property
    def concept_arguments(self) -> List[ConceptRef]:
        return self.function.concept_arguments + [x.reference for x in self.by]

    @property
    def output_datatype(self):
        return self.function.output_datatype

    @property
    def output_purpose(self):
        return self.function.output_purpose

    def with_merge(self, source: Concept, target: Concept, modifiers: List[Modifier]):
        return AggregateWrapper.model_construct(
            function=self.function.with_merge(source, target, modifiers=modifiers),
            by=(
                [c.with_merge(source, target, modifiers) for c in self.by]
                if self.by
                else []
            ),
        )

    def with_reference_replacement(self, source, target):
        return AggregateWrapper.model_construct(
            function=self.function.with_reference_replacement(source, target),
            by=(
                [c.with_reference_replacement(source, target) for c in self.by]
                if self.by
                else []
            ),
        )

    def with_namespace(self, namespace: str) -> "AggregateWrapper":
        return AggregateWrapper.model_construct(
            function=self.function.with_namespace(namespace),
            by=[c.with_namespace(namespace) for c in self.by] if self.by else [],
        )


class FilterItem(DataTyped, Namespaced, ConceptArgs, BaseModel):
    content: FuncArgs
    where: "WhereClause"

    @field_validator("content", mode="before")
    def enforce_concept_ref(cls, v):
        if isinstance(v, Concept):
            return ConceptRef(address=v.address, datatype=v.datatype)
        return v

    def __str__(self):
        return f"<Filter: {str(self.content)} where {str(self.where)}>"

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> "FilterItem":
        return FilterItem.model_construct(
            content=(
                self.content.with_merge(source, target, modifiers)
                if isinstance(self.content, Mergeable)
                else self.content
            ),
            where=self.where.with_merge(source, target, modifiers),
        )

    def with_namespace(self, namespace: str) -> "FilterItem":
        return FilterItem.model_construct(
            content=(
                self.content.with_namespace(namespace)
                if isinstance(self.content, Namespaced)
                else self.content
            ),
            where=self.where.with_namespace(namespace),
        )

    @property
    def output_datatype(self):
        return arg_to_datatype(self.content)

    @property
    def concept_arguments(self):
        if isinstance(self.content, ConceptRef):
            return [self.content] + self.where.concept_arguments
        elif isinstance(self.content, ConceptArgs):
            return self.content.concept_arguments + self.where.concept_arguments
        return self.where.concept_arguments


class RowsetLineage(Namespaced, Mergeable, BaseModel):
    name: str
    derived_concepts: List[ConceptRef]
    select: SelectLineage | MultiSelectLineage

    def with_namespace(self, namespace: str):
        return RowsetLineage.model_construct(
            name=self.name,
            derived_concepts=[
                x.with_namespace(namespace) for x in self.derived_concepts
            ],
            select=self.select.with_namespace(namespace),
        )

    def with_merge(self, source: Concept, target: Concept, modifiers: List[Modifier]):
        return RowsetLineage.model_construct(
            name=self.name,
            derived_concepts=[
                x.with_merge(source, target, modifiers) for x in self.derived_concepts
            ],
            select=self.select.with_merge(source, target, modifiers),
        )


class RowsetItem(Mergeable, DataTyped, ConceptArgs, Namespaced, BaseModel):
    content: ConceptRef
    rowset: RowsetLineage

    def __repr__(self):
        return f"<Rowset<{self.rowset.name}>: {str(self.content)}>"

    def __str__(self):
        return self.__repr__()

    def with_merge(self, source: Concept, target: Concept, modifiers: List[Modifier]):
        return RowsetItem.model_construct(
            content=self.content.with_merge(source, target, modifiers),
            rowset=self.rowset,
        )

    def with_namespace(self, namespace: str) -> "RowsetItem":
        return RowsetItem.model_construct(
            content=self.content.with_namespace(namespace),
            rowset=self.rowset.with_namespace(namespace),
        )

    @property
    def output(self) -> ConceptRef:
        return self.content

    @property
    def output_datatype(self):
        return self.content.datatype

    @property
    def concept_arguments(self):
        return [self.content]


class OrderBy(Mergeable, Namespaced, BaseModel):
    items: List[OrderItem]

    def with_namespace(self, namespace: str) -> "OrderBy":
        return OrderBy.model_construct(
            items=[x.with_namespace(namespace) for x in self.items]
        )

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> "OrderBy":
        return OrderBy.model_construct(
            items=[x.with_merge(source, target, modifiers) for x in self.items]
        )

    @property
    def concept_arguments(self):
        base = []
        for x in self.items:
            base += x.concept_arguments
        return base


class AlignClause(Namespaced, BaseModel):
    items: List[AlignItem]

    def with_namespace(self, namespace: str) -> "AlignClause":
        return AlignClause.model_construct(
            items=[x.with_namespace(namespace) for x in self.items]
        )


class DeriveItem(Namespaced, DataTyped, ConceptArgs, Mergeable, BaseModel):
    expr: Expr
    name: str
    namespace: str

    @property
    def derived_concept(self) -> str:
        return f"{self.namespace}.{self.name}"
        # return ConceptRef(
        #     address=f"{self.namespace}.{self.name}",
        #     datatype=arg_to_datatype(self.expr),
        # )

    def with_namespace(self, namespace):
        return DeriveItem.model_construct(
            expr=(self.expr.with_namespace(namespace) if self.expr else None),
            name=self.name,
            namespace=namespace,
        )

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> "DeriveItem":
        return DeriveItem.model_construct(
            expr=(
                self.expr.with_merge(source, target, modifiers)
                if isinstance(self.expr, Mergeable)
                else self.expr
            ),
            name=self.name,
            namespace=self.namespace,
        )

    def with_reference_replacement(self, source, target):
        return DeriveItem.model_construct(
            expr=(
                self.expr.with_reference_replacement(source, target)
                if isinstance(self.expr, Mergeable)
                else self.expr
            ),
            name=self.name,
            namespace=self.namespace,
        )


class DeriveClause(Mergeable, Namespaced, BaseModel):
    items: List[DeriveItem]

    def with_namespace(self, namespace: str) -> "DeriveClause":
        return DeriveClause.model_construct(
            items=[
                x.with_namespace(namespace) if isinstance(x, Namespaced) else x
                for x in self.items
            ]
        )

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> "DeriveClause":
        return DeriveClause.model_construct(
            items=[
                (
                    x.with_merge(source, target, modifiers)
                    if isinstance(x, Mergeable)
                    else x
                )
                for x in self.items
            ]
        )

    def with_reference_replacement(self, source, target):
        return DeriveClause.model_construct(
            items=[
                (
                    x.with_reference_replacement(source, target)
                    if isinstance(x, Mergeable)
                    else x
                )
                for x in self.items
            ]
        )


class SelectLineage(Mergeable, Namespaced, BaseModel):
    selection: List[ConceptRef]
    hidden_components: set[str]
    local_concepts: dict[str, Concept]
    order_by: Optional[OrderBy] = None
    limit: Optional[int] = None
    meta: Metadata = Field(default_factory=lambda: Metadata())
    grain: Grain = Field(default_factory=Grain)
    where_clause: Union["WhereClause", None] = Field(default=None)
    having_clause: Union["HavingClause", None] = Field(default=None)

    @property
    def output_components(self) -> List[ConceptRef]:
        return self.selection

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> SelectLineage:
        return SelectLineage.model_construct(
            selection=[x.with_merge(source, target, modifiers) for x in self.selection],
            hidden_components=self.hidden_components,
            local_concepts={
                x: y.with_merge(source, target, modifiers)
                for x, y in self.local_concepts.items()
            },
            order_by=(
                self.order_by.with_merge(source, target, modifiers)
                if self.order_by
                else None
            ),
            limit=self.limit,
            grain=self.grain.with_merge(source, target, modifiers),
            where_clause=(
                self.where_clause.with_merge(source, target, modifiers)
                if self.where_clause
                else None
            ),
            having_clause=(
                self.having_clause.with_merge(source, target, modifiers)
                if self.having_clause
                else None
            ),
        )

    def with_namespace(self, namespace):
        return SelectLineage.model_construct(
            selection=[x.with_namespace(namespace) for x in self.selection],
            hidden_components=self.hidden_components,
            local_concepts={
                x: y.with_namespace(namespace) for x, y in self.local_concepts.items()
            },
            order_by=self.order_by.with_namespace(namespace) if self.order_by else None,
            limit=self.limit,
            meta=self.meta,
            grain=self.grain.with_namespace(namespace),
            where_clause=(
                self.where_clause.with_namespace(namespace)
                if self.where_clause
                else None
            ),
            having_clause=(
                self.having_clause.with_namespace(namespace)
                if self.having_clause
                else None
            ),
        )


class MultiSelectLineage(Mergeable, ConceptArgs, Namespaced, BaseModel):
    selects: List[SelectLineage]
    align: AlignClause

    namespace: str
    order_by: Optional[OrderBy] = None
    limit: Optional[int] = None
    where_clause: Union["WhereClause", None] = Field(default=None)
    having_clause: Union["HavingClause", None] = Field(default=None)
    derive: DeriveClause | None = None
    hidden_components: set[str]

    @property
    def grain(self):
        base = Grain()
        for select in self.selects:
            base += select.grain
        return base

    @property
    def output_components(self) -> list[ConceptRef]:
        output = [
            ConceptRef.model_construct(address=x, datatype=DataType.UNKNOWN)
            for x in self.derived_concepts
        ]
        for select in self.selects:
            output += select.output_components
        return [x for x in output if x.address not in self.hidden_components]

    def with_merge(
        self, source: Concept, target: Concept, modifiers: List[Modifier]
    ) -> MultiSelectLineage:
        new = MultiSelectLineage.model_construct(
            selects=[s.with_merge(source, target, modifiers) for s in self.selects],
            align=self.align,
            derive=(
                self.derive.with_merge(source, target, modifiers)
                if self.derive
                else None
            ),
            namespace=self.namespace,
            hidden_components=self.hidden_components,
            order_by=(
                self.order_by.with_merge(source, target, modifiers)
                if self.order_by
                else None
            ),
            limit=self.limit,
            where_clause=(
                self.where_clause.with_merge(source, target, modifiers)
                if self.where_clause
                else None
            ),
            having_clause=(
                self.having_clause.with_merge(source, target, modifiers)
                if self.having_clause
                else None
            ),
        )
        return new

    def with_namespace(self, namespace: str) -> "MultiSelectLineage":
        return MultiSelectLineage.model_construct(
            selects=[c.with_namespace(namespace) for c in self.selects],
            align=self.align.with_namespace(namespace),
            derive=self.derive.with_namespace(namespace) if self.derive else None,
            namespace=namespace,
            hidden_components=self.hidden_components,
            order_by=self.order_by.with_namespace(namespace) if self.order_by else None,
            limit=self.limit,
            where_clause=(
                self.where_clause.with_namespace(namespace)
                if self.where_clause
                else None
            ),
            having_clause=(
                self.having_clause.with_namespace(namespace)
                if self.having_clause
                else None
            ),
        )

    @property
    def derived_concepts(self) -> set[str]:
        output = set()
        for item in self.align.items:
            output.add(item.aligned_concept)
        if self.derive:
            for ditem in self.derive.items:
                output.add(ditem.derived_concept)
        return output

    @property
    def concept_arguments(self):
        output = []
        for select in self.selects:
            output += select.output_components
        return unique(output, "address")


class LooseConceptList(BaseModel):
    concepts: Sequence[Concept | ConceptRef]

    @cached_property
    def addresses(self) -> set[str]:
        return {s.address for s in self.concepts}

    @classmethod
    def validate(cls, v):
        return cls(v)

    @cached_property
    def sorted_addresses(self) -> List[str]:
        return sorted(list(self.addresses))

    def __str__(self) -> str:
        return f"lcl{str(self.sorted_addresses)}"

    def __iter__(self):
        return iter(self.concepts)

    def __eq__(self, other):
        if not isinstance(other, LooseConceptList):
            return False
        return self.addresses == other.addresses

    def issubset(self, other):
        if not isinstance(other, LooseConceptList):
            return False
        return self.addresses.issubset(other.addresses)

    def __contains__(self, other):
        if isinstance(other, str):
            return other in self.addresses
        if not isinstance(other, Concept):
            return False
        return other.address in self.addresses

    def difference(self, other):
        if not isinstance(other, LooseConceptList):
            return False
        return self.addresses.difference(other.addresses)

    def isdisjoint(self, other):
        if not isinstance(other, LooseConceptList):
            return False
        return self.addresses.isdisjoint(other.addresses)


class AlignItem(Namespaced, BaseModel):
    alias: str
    concepts: List[ConceptRef]
    namespace: str = Field(default=DEFAULT_NAMESPACE, validate_default=True)

    @field_validator("concepts", mode="before")
    @classmethod
    def enforce_concept_ref(cls, v):
        output = []
        for item in v:
            if isinstance(item, Concept):
                output.append(item.reference)
            else:
                output.append(item)
        return output

    @computed_field  # type: ignore
    @cached_property
    def concepts_lcl(self) -> LooseConceptList:
        return LooseConceptList(concepts=self.concepts)

    @property
    def aligned_concept(self) -> str:
        return f"{self.namespace}.{self.alias}"

    def with_namespace(self, namespace: str) -> "AlignItem":
        return AlignItem.model_construct(
            alias=self.alias,
            concepts=[c.with_namespace(namespace) for c in self.concepts],
            namespace=namespace,
        )


class CustomFunctionFactory:
    def __init__(
        self,
        function: Expr,
        namespace: str,
        function_arguments: list[ArgBinding],
        name: str,
    ):
        self.namespace = namespace
        self.function = function
        self.function_arguments = function_arguments
        self.name = name

    def with_namespace(self, namespace: str):
        self.namespace = namespace
        self.function = (
            self.function.with_namespace(namespace)
            if isinstance(self.function, Namespaced)
            else self.function
        )
        self.function_arguments = [
            x.with_namespace(namespace) for x in self.function_arguments
        ]
        return self

    def __call__(self, *creation_args: ArgBinding | Expr):
        nout = (
            self.function.model_copy(deep=True)
            if isinstance(self.function, BaseModel)
            else self.function
        )
        creation_arg_list: list[ArgBinding | Expr] = list(creation_args)
        if len(creation_args) < len(self.function_arguments):
            for binding in self.function_arguments[len(creation_arg_list) :]:
                if binding.default is None:
                    raise ValueError(f"Missing argument {binding.name}")

                creation_arg_list.append(binding.default)
        for arg_idx, arg in enumerate(self.function_arguments):
            if not arg.datatype or arg.datatype == DataType.UNKNOWN:
                continue
            if arg_idx > len(creation_arg_list):
                continue
            comparison = arg_to_datatype(creation_arg_list[arg_idx])
            if comparison != arg.datatype:
                raise TypeError(
                    f"Invalid type passed into custom function @{self.name} in position {arg_idx+1} for argument {arg.name}, expected {arg.datatype}, got {comparison}"
                )
            if isinstance(arg.datatype, TraitDataType):
                if not (
                    isinstance(comparison, TraitDataType)
                    and all(x in comparison.traits for x in arg.datatype.traits)
                ):
                    raise TypeError(
                        f"Invalid argument type passed into custom function @{self.name} in position {arg_idx+1} for argument {arg.name}, expected traits {arg.datatype.traits}, got {comparison}"
                    )

        if isinstance(nout, Mergeable):
            for idx, x in enumerate(creation_arg_list):
                if self.namespace == DEFAULT_NAMESPACE:
                    target = f"{DEFAULT_NAMESPACE}.{self.function_arguments[idx].name}"
                else:
                    target = self.function_arguments[idx].name
                nout = nout.with_reference_replacement(target, x)
        return nout


class Metadata(BaseModel):
    """Metadata container object.
    TODO: support arbitrary tags"""

    description: Optional[str] = None
    line_number: Optional[int] = None
    concept_source: ConceptSource = ConceptSource.MANUAL


class Window(BaseModel):
    count: int
    window_order: WindowOrder

    def __str__(self):
        return f"Window<{self.window_order}>"


class WindowItemOver(BaseModel):
    contents: List[ConceptRef]


class WindowItemOrder(BaseModel):
    contents: List["OrderItem"]


class Comment(BaseModel):
    text: str


class ArgBinding(Namespaced, DataTyped, BaseModel):
    name: str
    default: Expr | None = None
    datatype: (
        DataType | MapType | ArrayType | NumericType | StructType | TraitDataType
    ) = DataType.UNKNOWN

    def with_namespace(self, namespace):
        return ArgBinding(
            name=address_with_namespace(self.name, namespace),
            default=(
                self.default.with_namespace(namespace)
                if isinstance(self.default, Namespaced)
                else self.default
            ),
        )

    @property
    def output_datatype(self):
        if self.default is not None:
            return arg_to_datatype(self.default)
        return self.datatype


class CustomType(BaseModel):
    name: str
    type: DataType | list[DataType]
    drop_on: list[FunctionType] = Field(default_factory=list)
    add_on: list[FunctionType] = Field(default_factory=list)

    def with_namespace(self, namespace: str) -> "CustomType":
        return CustomType.model_construct(
            name=address_with_namespace(self.name, namespace),
            type=self.type,
            drop_on=self.drop_on,
            add_on=self.add_on,
        )


Expr = (
    MagicConstants
    | bool
    | int
    | str
    | float
    | date
    | datetime
    | TupleWrapper
    | ListWrapper
    | MapWrapper
    | WindowItem
    | FilterItem
    | ConceptRef
    | Comparison
    | Conditional
    | FunctionCallWrapper
    | Parenthetical
    | Function
    | AggregateWrapper
    | CaseWhen
    | CaseElse
)

FuncArgs = (
    ConceptRef
    | AggregateWrapper
    | Function
    | FunctionCallWrapper
    | Parenthetical
    | CaseWhen
    | CaseElse
    | WindowItem
    | FilterItem
    | bool
    | int
    | float
    | DatePart
    | str
    | date
    | datetime
    | MapWrapper[Any, Any]
    | TraitDataType
    | DataType
    | ArrayType
    | MapType
    | NumericType
    | ListWrapper[Any]
    | TupleWrapper[Any]
    | Comparison
    | Conditional
    | MagicConstants
    | ArgBinding
    | Ordering
)



from dataclasses import dataclass
from typing import Any, List, Sequence

from trilogy.core.enums import Modifier
from trilogy.core.models.core import (
    ArrayType,
    DataType,
    MapType,
    NumericType,
    StructType,
    TraitDataType,
)


class UndefinedConceptException(Exception):
    def __init__(self, message, suggestions: List[str]):
        super().__init__(self, message)
        self.message = message
        self.suggestions = suggestions


class FrozenEnvironmentException(Exception):
    pass


class InvalidSyntaxException(Exception):
    pass


class UnresolvableQueryException(Exception):
    pass


class NoDatasourceException(UnresolvableQueryException):
    pass


class ModelValidationError(Exception):
    def __init__(
        self,
        message,
        children: Sequence["ModelValidationError"] | None = None,
        **kwargs,
    ):
        super().__init__(self, message, **kwargs)
        self.message = message
        self.children = children


class DatasourceModelValidationError(ModelValidationError):
    pass


class DatasourceGrainValidationError(DatasourceModelValidationError):
    pass


@dataclass
class DatasourceColumnBindingData:
    address: str
    value: Any
    value_type: (
        DataType | ArrayType | StructType | MapType | NumericType | TraitDataType
    )
    value_modifiers: List[Modifier]
    actual_type: (
        DataType | ArrayType | StructType | MapType | NumericType | TraitDataType
    )
    actual_modifiers: List[Modifier]

    def format_failure(self):
        return f"Concept {self.address} value '{self.value}' with type {self.value_type} and {self.value_modifiers} does not conform to expected type {str(self.actual_type)} with modifiers {self.actual_modifiers}"

    def is_modifier_issue(self) -> bool:
        return len(self.value_modifiers) > 0 and any(
            [x not in self.actual_modifiers for x in self.value_modifiers]
        )

    def is_type_issue(self) -> bool:
        return self.value_type != self.actual_type


class DatasourceColumnBindingError(DatasourceModelValidationError):
    def __init__(
        self,
        address: str,
        errors: list[DatasourceColumnBindingData],
        message: str | None = None,
    ):
        if not message:
            message = f"Datasource {address} failed validation. Found rows that do not conform to types: {[failure.format_failure() for failure in errors]}"
        super().__init__(message)
        self.errors = errors
        self.dataset_address = address


class ConceptModelValidationError(ModelValidationError):
    pass


class AmbiguousRelationshipResolutionException(UnresolvableQueryException):
    def __init__(self, message, parents: List[set[str]]):
        super().__init__(self, message)
        self.message = message
        self.parents = parents



from trilogy.core.constants import ALL_ROWS_CONCEPT, INTERNAL_NAMESPACE
from trilogy.core.models.build import BuildGrain
from trilogy.core.models.environment import Environment
from trilogy.core.processing.concept_strategies_v3 import (
    History,
    search_concepts,
)
from trilogy.core.processing.node_generators import gen_group_node
from trilogy.core.processing.nodes import GroupNode


def test_group_node(test_environment, test_environment_graph):
    history = History(base_environment=test_environment)
    test_environment = test_environment.materialize_for_select()
    total_revenue = test_environment.concepts["total_revenue"]
    revenue = test_environment.concepts["revenue"]
    category = test_environment.concepts["category_name"]
    group_node = GroupNode(
        output_concepts=[total_revenue, category],
        input_concepts=[category, revenue],
        environment=test_environment,
        parents=[
            search_concepts(
                [category, revenue],
                environment=test_environment,
                g=test_environment_graph,
                depth=0,
                history=history,
            )
        ],
    )
    group_node.resolve()


def test_group_node_property(test_environment: Environment, test_environment_graph):
    history = History(base_environment=test_environment)
    test_environment = test_environment.materialize_for_select()
    sum_name_length = test_environment.concepts["category_name_length_sum"]

    group_node = gen_group_node(
        sum_name_length,
        local_optional=[],
        environment=test_environment,
        g=test_environment_graph,
        source_concepts=search_concepts,
        history=history,
        depth=0,
    )
    input_concept_names = {
        x.name
        for x in group_node.parents[0].output_concepts
        if x not in group_node.parents[0].hidden_concepts
    }
    assert input_concept_names == {
        "category_name_length",
        # "category_name",
        "category_id",
    }
    final = group_node.resolve()
    assert len(final.datasources) == 1
    assert final.datasources[0].group_required is False


def test_group_node_property_all(test_environment: Environment, test_environment_graph):
    history = History(base_environment=test_environment)
    test_environment = test_environment.materialize_for_select()
    sum_name_length = test_environment.concepts["category_name_length_sum"]
    all_rows = test_environment.concepts[f"{INTERNAL_NAMESPACE}.{ALL_ROWS_CONCEPT}"]
    sum_name_length_all_rows = sum_name_length.with_grain(
        BuildGrain(components={all_rows.address})
    )
    group_node = gen_group_node(
        sum_name_length_all_rows,
        local_optional=[],
        environment=test_environment,
        g=test_environment_graph,
        source_concepts=search_concepts,
        history=history,
        depth=0,
    )
    input_concept_names = {
        x.name
        for x in group_node.parents[0].output_concepts
        if x not in group_node.parents[0].hidden_concepts
    }
    assert input_concept_names == {
        "category_name_length",
        # "category_name",
        "category_id",
    }
    final = group_node.resolve()
    assert len(final.datasources) == 1
    assert final.datasources[0].group_required is False



from dataclasses import dataclass
from typing import Literal, Optional


@dataclass
class UsageDict:
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


@dataclass
class LLMResponse:
    text: str
    usage: UsageDict


@dataclass
class LLMRequestOptions:
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None


@dataclass
class LLMMessage:
    role: Literal["user", "assistant", "system"]
    content: str
    model_info: Optional[dict] = None
    hidden: bool = False  # Used to hide messages in the UI

    def __post_init__(self):
        if self.model_info is None:
            self.model_info = {}



from typing import List

from trilogy.core.models.build import BuildConcept, BuildWhereClause
from trilogy.core.models.build_environment import BuildEnvironment
from trilogy.core.processing.nodes import History, StrategyNode

LOGGER_PREFIX = "[GEN_CONSTANT_NODE]"


def gen_constant_node(
    concept: BuildConcept,
    local_optional: List[BuildConcept],
    environment: BuildEnvironment,
    g,
    depth: int,
    source_concepts,
    history: History | None = None,
    conditions: BuildWhereClause | None = None,
    accept_partial: bool = False,
):
    """our only goal here is to generate a row if conditions exist, or none if they do not"""

    targets = [concept] + local_optional
    if conditions:
        targets += conditions.row_arguments
    parent_node: StrategyNode | None = source_concepts(
        mandatory_list=targets,
        environment=environment,
        g=g,
        depth=depth + 1,
        history=history,
        conditions=conditions,
        accept_partial=accept_partial,
    )
    if not parent_node:
        return None
    parent_node.set_output_concepts([concept] + local_optional)
    return parent_node



from functools import cached_property
from pathlib import Path
from typing import Annotated, List, Optional, Union

from pydantic import BaseModel, Field, computed_field, field_validator
from pydantic.functional_validators import PlainValidator

from trilogy.constants import CONFIG, DEFAULT_NAMESPACE
from trilogy.core.enums import (
    ConceptSource,
    FunctionClass,
    IOType,
    Modifier,
    ShowCategory,
    ValidationScope,
)
from trilogy.core.models.author import (
    AggregateWrapper,
    AlignClause,
    ArgBinding,
    Concept,
    ConceptRef,
    CustomType,
    DeriveClause,
    Expr,
    FilterItem,
    Function,
    FunctionCallWrapper,
    Grain,
    HasUUID,
    HavingClause,
    Metadata,
    MultiSelectLineage,
    OrderBy,
    Parenthetical,
    SelectLineage,
    UndefinedConcept,
    WhereClause,
    WindowItem,
)
from trilogy.core.models.datasource import Address, ColumnAssignment, Datasource
from trilogy.core.models.environment import (
    Environment,
    EnvironmentConceptDict,
    validate_concepts,
)
from trilogy.core.statements.common import SelectTypeMixin
from trilogy.utility import unique


class ConceptTransform(BaseModel):
    function: (
        Function
        | FilterItem
        | WindowItem
        | AggregateWrapper
        | FunctionCallWrapper
        | Parenthetical
    )
    output: Concept  # this has to be a full concept, as it may not exist in environment
    modifiers: List[Modifier] = Field(default_factory=list)

    def with_merge(self, source: Concept, target: Concept, modifiers: List[Modifier]):
        return ConceptTransform(
            function=self.function.with_merge(source, target, modifiers),
            output=self.output.with_merge(source, target, modifiers),
            modifiers=self.modifiers + modifiers,
        )

    def with_namespace(self, namespace: str) -> "ConceptTransform":
        return ConceptTransform(
            function=self.function.with_namespace(namespace),
            output=self.output.with_namespace(namespace),
            modifiers=self.modifiers,
        )


class SelectItem(BaseModel):
    content: Union[ConceptTransform, ConceptRef]
    modifiers: List[Modifier] = Field(default_factory=list)

    @field_validator("content", mode="before")
    def parse_content(cls, v):
        if isinstance(v, Concept):
            return v.reference
        return v

    @property
    def concept(self) -> ConceptRef:
        if isinstance(self.content, (ConceptRef)):
            return self.content
        elif isinstance(self.content, Concept):
            return self.content.reference
        return self.content.output.reference

    @property
    def is_undefined(self) -> bool:
        return True if isinstance(self.content, UndefinedConcept) else False


class SelectStatement(HasUUID, SelectTypeMixin, BaseModel):
    selection: List[SelectItem]
    order_by: Optional[OrderBy] = None
    limit: Optional[int] = None
    meta: Metadata = Field(default_factory=lambda: Metadata())
    local_concepts: Annotated[
        EnvironmentConceptDict, PlainValidator(validate_concepts)
    ] = Field(default_factory=EnvironmentConceptDict)
    grain: Grain = Field(default_factory=Grain)

    def as_lineage(self, environment: Environment) -> SelectLineage:
        return SelectLineage(
            selection=[
                environment.concepts[x.concept.address].reference
                for x in self.selection
            ],
            order_by=self.order_by,
            limit=self.limit,
            where_clause=self.where_clause,
            having_clause=self.having_clause,
            local_concepts=self.local_concepts,
            hidden_components=self.hidden_components,
            grain=self.grain,
        )

    @classmethod
    def from_inputs(
        cls,
        environment: Environment,
        selection: List[SelectItem],
        order_by: OrderBy | None = None,
        limit: int | None = None,
        meta: Metadata | None = None,
        where_clause: WhereClause | None = None,
        having_clause: HavingClause | None = None,
    ) -> "SelectStatement":
        output = SelectStatement(
            selection=selection,
            where_clause=where_clause,
            having_clause=having_clause,
            limit=limit,
            order_by=order_by,
            meta=meta or Metadata(),
        )

        output.grain = output.calculate_grain(environment, output.local_concepts)

        for x in selection:
            if x.is_undefined and environment.concepts.fail_on_missing:
                environment.concepts.raise_undefined(
                    x.concept.address, meta.line_number if meta else None
                )
            elif isinstance(x.content, ConceptTransform):
                if isinstance(x.content.output, UndefinedConcept):
                    continue
                if CONFIG.parsing.select_as_definition and not environment.frozen:
                    if x.concept.address not in environment.concepts:
                        environment.add_concept(x.content.output, add_derived=False)
                    elif x.concept.address in environment.concepts:
                        version = environment.concepts[x.concept.address]
                        if version.metadata.concept_source == ConceptSource.SELECT:
                            environment.add_concept(
                                x.content.output, force=True, add_derived=False
                            )
                x.content.output = x.content.output.set_select_grain(
                    output.grain, environment
                )
                # we might not need this
                output.local_concepts[x.content.output.address] = x.content.output

            elif isinstance(x.content, ConceptRef):
                output.local_concepts[x.content.address] = environment.concepts[
                    x.content.address
                ]
        output.grain = output.calculate_grain(environment, output.local_concepts)
        output.validate_syntax(environment)
        return output

    def calculate_grain(
        self,
        environment: Environment | None = None,
        local_concepts: dict[str, Concept] | None = None,
    ) -> Grain:
        targets = []
        for x in self.selection:
            targets.append(x.concept)

        result = Grain.from_concepts(
            targets,
            where_clause=self.where_clause,
            environment=environment,
            local_concepts=local_concepts,
        )
        return result

    def validate_syntax(self, environment: Environment):
        if self.where_clause:
            for x in self.where_clause.concept_arguments:
                if isinstance(x, UndefinedConcept):
                    validate = environment.concepts.get(x.address)
                    if validate and self.where_clause:
                        self.where_clause = (
                            self.where_clause.with_reference_replacement(
                                x.address, validate.reference
                            )
                        )
                    else:
                        environment.concepts.raise_undefined(
                            x.address, x.metadata.line_number if x.metadata else None
                        )
        all_in_output = [x for x in self.output_components]
        if self.where_clause:
            for cref in self.where_clause.concept_arguments:
                concept = environment.concepts[cref.address]
                if isinstance(concept, UndefinedConcept):
                    continue
                if (
                    concept.lineage
                    and isinstance(concept.lineage, Function)
                    and concept.lineage.operator
                    in FunctionClass.AGGREGATE_FUNCTIONS.value
                ):
                    if concept.address in self.locally_derived:
                        raise SyntaxError(
                            f"Cannot reference an aggregate derived in the select ({concept.address}) in the same statement where clause; move to the HAVING clause instead; Line: {self.meta.line_number}"
                        )

                if (
                    concept.lineage
                    and isinstance(concept.lineage, AggregateWrapper)
                    and concept.lineage.function.operator
                    in FunctionClass.AGGREGATE_FUNCTIONS.value
                ):
                    if concept.address in self.locally_derived:
                        raise SyntaxError(
                            f"Cannot reference an aggregate derived in the select ({concept.address}) in the same statement where clause; move to the HAVING clause instead; Line: {self.meta.line_number}"
                        )
        if self.having_clause:
            for cref in self.having_clause.concept_arguments:
                if cref.address not in [x for x in self.output_components]:
                    raise SyntaxError(
                        f"Cannot reference a column ({cref.address}) that is not in the select projection in the HAVING clause, move to WHERE;  Line: {self.meta.line_number}"
                    )
        if self.order_by:
            for cref in self.order_by.concept_arguments:
                if cref.address not in all_in_output:
                    raise SyntaxError(
                        f"Cannot order by column {cref.address} that is not in the output projection; line: {self.meta.line_number}"
                    )

    def __str__(self):
        from trilogy.parsing.render import render_query

        return render_query(self)

    @field_validator("selection", mode="before")
    @classmethod
    def selection_validation(cls, v):
        new = []
        for item in v:
            if isinstance(item, (Concept, ConceptTransform)):
                new.append(SelectItem(content=item))
            else:
                new.append(item)
        return new

    @property
    def locally_derived(self) -> set[str]:
        locally_derived: set[str] = set()
        for item in self.selection:
            if isinstance(item.content, ConceptTransform):
                locally_derived.add(item.concept.address)
        return locally_derived

    @property
    def output_components(self) -> List[ConceptRef]:
        return [x.concept for x in self.selection]

    @property
    def hidden_components(self) -> set[str]:
        return set(
            x.concept.address for x in self.selection if Modifier.HIDDEN in x.modifiers
        )

    def to_datasource(
        self,
        namespace: str,
        name: str,
        address: Address,
        environment: Environment,
        grain: Grain | None = None,
    ) -> Datasource:
        if self.where_clause or self.having_clause:
            modifiers = [Modifier.PARTIAL]
        else:
            modifiers = []
        columns = [
            # TODO: replace hardcoded replacement here
            # if the concept is a locally derived concept, it cannot ever be partial
            # but if it's a concept pulled in from upstream and we have a where clause, it should be partial
            ColumnAssignment(
                alias=(
                    c.address.replace(".", "_")
                    if c.namespace != DEFAULT_NAMESPACE
                    else c.name
                ),
                concept=environment.concepts[c.address].reference,
                modifiers=modifiers if c.address not in self.locally_derived else [],
            )
            for c in self.output_components
        ]

        condition = None
        if self.where_clause:
            condition = self.where_clause.conditional
        if self.having_clause:
            if condition:
                condition = self.having_clause.conditional + condition
            else:
                condition = self.having_clause.conditional

        new_datasource = Datasource(
            name=name,
            address=address,
            grain=grain or self.grain,
            columns=columns,
            namespace=namespace,
            non_partial_for=WhereClause(conditional=condition) if condition else None,
        )
        return new_datasource


class RawSQLStatement(BaseModel):
    text: str
    meta: Optional[Metadata] = Field(default_factory=lambda: Metadata())


class CopyStatement(BaseModel):
    target: str
    target_type: IOType
    meta: Optional[Metadata] = Field(default_factory=lambda: Metadata())
    select: SelectStatement


class MultiSelectStatement(HasUUID, SelectTypeMixin, BaseModel):
    selects: List[SelectStatement]
    align: AlignClause
    namespace: str
    derived_concepts: List[Concept]
    order_by: Optional[OrderBy] = None
    limit: Optional[int] = None
    meta: Optional[Metadata] = Field(default_factory=lambda: Metadata())
    local_concepts: Annotated[
        EnvironmentConceptDict, PlainValidator(validate_concepts)
    ] = Field(default_factory=EnvironmentConceptDict)
    derive: DeriveClause | None = None

    def as_lineage(self, environment: Environment):
        return MultiSelectLineage(
            selects=[x.as_lineage(environment) for x in self.selects],
            align=self.align,
            derive=self.derive,
            namespace=self.namespace,
            # derived_concepts = self.derived_concepts,
            limit=self.limit,
            order_by=self.order_by,
            where_clause=self.where_clause,
            having_clause=self.having_clause,
            hidden_components=self.hidden_components,
        )

    def __repr__(self):
        return "MultiSelect<" + " MERGE ".join([str(s) for s in self.selects]) + ">"

    @property
    def grain(self):
        base = Grain()
        for select in self.selects:
            base += select.grain
        return base

    @property
    def output_components(self) -> List[ConceptRef]:
        output = [x.reference for x in self.derived_concepts]
        for select in self.selects:
            output += [
                x
                for x in select.output_components
                if x.address not in select.hidden_components
            ]
        return unique(output, "address")

    @computed_field  # type: ignore
    @cached_property
    def hidden_components(self) -> set[str]:
        output: set[str] = set()
        for select in self.selects:
            output = output.union(select.hidden_components)
        return output

    @property
    def locally_derived(self) -> set[str]:
        locally_derived: set[str] = set([x.address for x in self.derived_concepts])
        for select in self.selects:
            locally_derived = locally_derived.union(select.locally_derived)
        return locally_derived


class RowsetDerivationStatement(HasUUID, BaseModel):
    name: str
    select: SelectStatement | MultiSelectStatement
    namespace: str

    def __repr__(self):
        return f"RowsetDerivation<{str(self.select)}>"

    def __str__(self):
        return self.__repr__()


class MergeStatementV2(HasUUID, BaseModel):
    sources: list[Concept]
    targets: dict[str, Concept]
    source_wildcard: str | None = None
    target_wildcard: str | None = None
    modifiers: List[Modifier] = Field(default_factory=list)


class KeyMergeStatement(HasUUID, BaseModel):
    keys: set[str]
    target: ConceptRef


class ImportStatement(HasUUID, BaseModel):
    # import abc.def as bar
    # the bit after 'as', eg bar
    alias: str
    # the bit after import, abc.def
    input_path: str
    # what it actually resolves to, typically a filepath
    path: Path


class PersistStatement(HasUUID, BaseModel):
    datasource: Datasource
    select: SelectStatement
    meta: Optional[Metadata] = Field(default_factory=lambda: Metadata())

    @property
    def identifier(self):
        return self.datasource.identifier

    @property
    def address(self):
        return self.datasource.address


class ValidateStatement(BaseModel):
    scope: ValidationScope
    targets: Optional[List[str]] = None  # list of identifiers


class ShowStatement(BaseModel):
    content: SelectStatement | PersistStatement | ValidateStatement | ShowCategory


class Limit(BaseModel):
    count: int


class ConceptDeclarationStatement(HasUUID, BaseModel):
    concept: Concept


class ConceptDerivationStatement(BaseModel):
    concept: Concept


class TypeDeclaration(BaseModel):
    type: CustomType


class FunctionDeclaration(HasUUID, BaseModel):
    name: str
    args: list[ArgBinding]
    expr: Expr



from abc import ABC, abstractmethod
from typing import List, Optional

from trilogy.ai.enums import Provider
from trilogy.ai.models import LLMMessage, LLMRequestOptions, LLMResponse


class LLMProvider(ABC):
    def __init__(self, name: str, api_key: str, model: str, provider: Provider):
        self.api_key = api_key
        self.models: List[str] = []
        self.name = name
        self.model = model
        self.type = provider
        self.error: Optional[str] = None

    # Abstract method to be implemented by specific providers
    @abstractmethod
    def generate_completion(
        self, options: LLMRequestOptions, history: List[LLMMessage]
    ) -> LLMResponse:
        pass



from collections import defaultdict

from trilogy.constants import CONFIG
from trilogy.core.models.build import BuildDatasource
from trilogy.core.models.execute import CTE, RecursiveCTE, UnionCTE
from trilogy.core.optimizations.base_optimization import OptimizationRule


class InlineDatasource(OptimizationRule):
    def __init__(self):
        super().__init__()
        self.candidates = defaultdict(lambda: set())
        self.count = defaultdict(lambda: 0)

    def optimize(
        self, cte: CTE | UnionCTE, inverse_map: dict[str, list[CTE | UnionCTE]]
    ) -> bool:
        if isinstance(cte, UnionCTE):
            return any(
                self.optimize(x, inverse_map=inverse_map) for x in cte.internal_ctes
            )
        if isinstance(cte, RecursiveCTE):
            return False
        if not cte.parent_ctes:
            return False

        self.debug(
            f"Checking {cte.name} for consolidating inline tables with {len(cte.parent_ctes)} parents"
        )
        to_inline: list[CTE] = []
        force_group = False
        for parent_cte in cte.parent_ctes:
            if isinstance(parent_cte, UnionCTE):
                continue
            if isinstance(parent_cte, RecursiveCTE):
                continue
            if not parent_cte.is_root_datasource:
                self.debug(f"Cannot inline: parent {parent_cte.name} is not root")
                continue
            if parent_cte.parent_ctes:
                self.debug(f"Cannot inline: parent {parent_cte.name} has parents")
                continue
            if parent_cte.condition:
                self.debug(
                    f"Cannot inline: parent {parent_cte.name} has condition, cannot be inlined"
                )
                continue
            raw_root = parent_cte.source.datasources[0]
            if not isinstance(raw_root, BuildDatasource):
                self.debug(f"Cannot inline: Parent {parent_cte.name} is not datasource")
                continue
            root: BuildDatasource = raw_root
            if not root.can_be_inlined:
                self.debug(
                    f"Cannot inline: Parent {parent_cte.name} datasource is not inlineable"
                )
                continue
            root_outputs = {x.address for x in root.output_concepts}
            inherited = {
                x for x, v in cte.source_map.items() if v and parent_cte.name in v
            }
            if not inherited.issubset(root_outputs):
                cte_missing = inherited - root_outputs
                self.log(
                    f"Cannot inline: Not all required inputs to {parent_cte.name} are found on datasource, missing {cte_missing}"
                )
                continue
            if not root.grain.issubset(parent_cte.grain):
                self.log(
                    f"Cannot inline: {parent_cte.name} is at wrong grain to inline ({root.grain} vs {parent_cte.grain})"
                )
                continue
            to_inline.append(parent_cte)

        optimized = False
        for replaceable in to_inline:
            if replaceable.name not in self.candidates[cte.name]:
                self.candidates[cte.name].add(replaceable.name)
                self.count[replaceable.source.identifier] += 1
                return True
            if (
                self.count[replaceable.source.identifier]
                > CONFIG.optimizations.constant_inline_cutoff
            ):
                self.log(
                    f"Skipping inlining raw datasource {replaceable.source.identifier} ({replaceable.name}) due to multiple references"
                )
                continue
            if not replaceable.source.datasources[0].grain.issubset(replaceable.grain):
                self.log(
                    f"Forcing group ({parent_cte.grain} being replaced by inlined source {root.grain})"
                )
                force_group = True
            result = cte.inline_parent_datasource(replaceable, force_group=force_group)
            if result:
                self.log(
                    f"Inlined parent {replaceable.name} with {replaceable.source.identifier}"
                )
                optimized = True
            else:
                self.log(f"Failed to inline {replaceable.name}")
        return optimized
